//=============================================================================
//
//      cpuio.c
//
//      CPU/Board Specific IO
//
// Author(s):    Michael Kelly, Cogent Computer Systems, Inc.
// Date:         09/18/2005
// Description:  This file contains the IO functions required by Micro Monitor
//				 that are unique to the CSB535FS
// Modified for use with IMX21 evaluation board,  May 2006.
//
//
//=============================================================================

#include "config.h"
#include "cpuio.h"
#include "genlib.h"
#include "ether.h"
#include "stddefs.h"
#include "warmstart.h"
#include "mc9328mx21_uart.h"
#include "mc9328mx21.h"
#include "mc9328mx21_mem.h"
#include "cpu_gpio.h"
#include "umongpio.h"

extern ulong getpsr(void);
extern void putpsr(ulong);

// Delay for some usecs. - Not accurate, assumes ROM mode
// and no Cache
void udelay(int delay)
{
  volatile int i;
  for ( i = LOOPS_PER_USEC * delay; i ; i--)
	;
}

#define DIVISOR 10000

/* InitUART():
 * This is the top-level UART initialization for the monitor. 
 */
void
InitUART(int baud)
{
    if (baud == 0)
        baud = ConsoleBaudRate;

    /* Keep track of the last baud rate, so that it can be used if the */
    /* incoming baudrate is NULL. */
    ConsoleBaudRate = baud;

	// reset tx and rx
	UART1_REG(UART_CR2) = 0;

	// clear any pending interrupts and mask all further interrupts
	UART1_REG(UART_SR1) = 0xffffffff;
	UART1_REG(UART_SR2) = 0xffffffff;
	UART1_REG(UART_CR1) = 0x00000000;
	UART1_REG(UART_CR4) = 0x00000000;

	// set the RX mux to Uart, not IrDA - this is different than i.MXL
	UART1_REG(UART_CR3) = UART_CR3_RXDMUXSEL;

	// enable the uart and it's clocks - no auto baud, no interrupts
	UART1_REG(UART_CR1) |= UART_CR1_UARTEN;

	//enable the uart0 clock
	PLL_REG(PLL_PCCR0) = 0x00000001;

	// set the word parameters
	UART1_REG(UART_CR2) = UART_CR2_WS 	// 8-bit, 1 stop, no parity
					| UART_CR2_IRTS		// ignore RTS 
					| UART_CR2_SRST;	// no reset

	// set the fifo's to legal interrupt threshold values - unused by us
	// set input clock divisor to 1
	UART1_REG(UART_FCR) = UART_FCR_TXTL(2)
					| UART_FCR_RFDIV1
					| UART_FCR_RXTL(1);

	// set the baud rate denominator to 20,000.  Then divide
	// the desired baud by 100 and put that into the numerator.
	// This works when using a reference clock of 16Mhz with
	// RFDIV set to 1.  It also only works down to 1200 baud,
	// but who cares!
	UART1_REG(UART_BIR) = (baud/100) - 1;	// numerator
	UART1_REG(UART_BMR) = DIVISOR - 1;		// denominator

	// Enable TX and RX, release SRST
	UART1_REG(UART_CR2) |= UART_CR2_TXEN
					| UART_CR2_RXEN
					| UART_CR2_SRST;

    InitRemoteIO();
}

int
ConsoleBaudSet(int baud)
{
	/* Set baud rate denominator.  Then divide desired baud by
	 * 100 and put that into the numerator.  This works when
	 * using a reference clock of 16Mhz with RFDIV set to 1.
	 * It also only works down to 1200 baud, but who cares!
	 */
	UART1_REG(UART_BIR) = (baud/100) - 1;	// numerator
	UART1_REG(UART_BMR) = DIVISOR - 1;		// denominator
	return(0);
}

int
devInit(int baud)
{
	// disable all perpihperal clocks.  They have to enable themselves
	PLL_REG(PLL_PCCR0) = 0 | (1 << 17);

	// Disable all GPIO for now.  We will allow the gpio_init()
	// code to do it later.
	GPIOA_REG(GPIO_GIUS) = 0;	// no GPIO on Port A for now
	GPIOA_REG(GPIO_GPR) = 0;
	GPIOB_REG(GPIO_GIUS) = 0;	// no GPIO on Port B for now
	GPIOB_REG(GPIO_GPR) = 0;
	GPIOC_REG(GPIO_GIUS) = 0;	// no GPIO on Port C for now
	GPIOC_REG(GPIO_GPR) = 0;
	GPIOD_REG(GPIO_GIUS) = 0;	// no GPIO on Port D for now
	GPIOD_REG(GPIO_GPR) = 0;
	GPIOE_REG(GPIO_GIUS) = 0;	// no GPIO on Port E for now
	GPIOE_REG(GPIO_GPR) = 0;
	GPIOF_REG(GPIO_GIUS) = 0;	// no GPIO on Port F for now
	GPIOF_REG(GPIO_GPR) = 0;


        GPIOE_REG(GPIO_GIUS) |= 0x20000;
        GPIOE_REG(GPIO_OCR2) = 0xffffffff;
        GPIOE_REG(GPIO_DR)   |= 0x20000;
        GPIOE_REG(GPIO_DDIR) |= 0x20000;
/*
        GPIOB_REG(GPIO_GIUS) = (1 << 5);
        GPIOB_REG(GPIO_DDIR) = (1 << 5);
        GPIOB_REG(GPIO_OCR1) = 0xffffffff;
        GPIOB_REG(GPIO_DR)   = (1 << 5);
*/
	// Initialize GPIO
//        GPIO_init();	

    // Initialize the UART
    InitUART(baud);

	// Setup CS for IO
    EIM_REG(EIM_CS1U) =
		EIM_CSU_CNC_0		// hold CS low zero clks at the end of all accesses
	  | EIM_CSU_WSC(32)		// wait = 660ns (32+1 x 20ns, assumes BCLK = 48Mhz)
	  | EIM_CSU_WWS(0)		// No additional wait states for writes
	  | EIM_CSU_EDC(0);		// Zero extra dead cycles

	EIM_REG(EIM_CS1L) =
		EIM_CSL_OEA(1)		// CS asserted one 1/2 clocks before OE
	  | EIM_CSL_OEN(1)		// CS held one 1/2 clocks after OE
	  | EIM_CSL_WEA(1)		// CS asserted one 1/2 clocks before WE
	  | EIM_CSL_WEN(1)		// CS held one 1/2 clocks after WE
	  | EIM_CSL_CSA(8)		// setup and hold address 8 clocks around CS
//	  | EIM_CSL_EBC			// assert EB0-3 for writes only
	  | EIM_CSL_DSZ_HALF_0	// 16-bits on D0-15
//	  | EIM_CSL_DSZ_HALF_1	// 16-bits on D16-31
//	  | EIM_CSL_DSZ_WORD	// 32-bits on D0-15
	  | EIM_CSL_CSEN;		// enable it

	return 0;
}

int
target_console_empty(void)
{
	if (UART1_REG(UART_SR2) & UART_SR2_TXFE)
		return(1);
	return(0);
}

int
target_putchar(char c)
{
    /* Wait for transmit ready bit, then send character.
	 */
    while(!(UART1_REG(UART_SR2) & UART_SR2_TXFE)){}
	UART1_REG(UART_TXD) = (ulong)c;
    return((int)c);
}

int
target_gotachar(void)
{

    if (UART1_REG(UART_SR2) & UART_SR2_RDR)
        return(1);
    return(0);
}

int 
target_getchar(void)
{
    char    c;

    c = (uchar)UART1_REG(UART_RXD) & 0xff;
    return((int)c);
}

void 
target_reset(void)
{
	warmstart(INITIALIZE);
}

void
initCPUio()
{
}

void
intsrestore(psr)
ulong     psr;
{
    putpsr(psr);
}

/*
 * Read the program status register (CPSR)
 * and set the FIQ and IRQ bits.
 */
ulong
intsoff(void)
{
    ulong  psr;

    psr = getpsr();

    /*
     * Set bit 6, bit 7 to disable interrupts.
     */
    putpsr(psr | 0x000000c0);
    return(psr);
}

/* cacheInitForTarget():
    Enable instruction cache only...
*/
void
cacheInitForTarget()
{
    asm("   MRC p15, 0, r0, c1, c0, 0");
    asm("   ORR r0, r0, #0x1000");  /* bit 12 is ICACHE enable*/
    asm("   MCR p15, 0, r0, c1, c0, 0");

    /* Flush instruction cache */
    asm("   MCR p15, 0, r0, c7, c5, 0");
}
