#include "config.h"
#include "stddefs.h"
#include "cpuio.h"
#include "genlib.h"
#include "warmstart.h"

/* devInit():
 * As a bare minimum, initialize the console UART here using the
 * incoming 'baud' value as the baud rate.
 */
void
devInit(int baud)
{
  // disable all interrupt sources
  IOWR_UART_CONTROL( UART1_BASE, 0);

  // note the nios uart may have a fixed baud rate
  IOWR_UART_DIVISOR( UART1_BASE, (((( UART1_FREQ * 10) / UART1_BAUD) + 5) / 10));

}

/* ConsoleBaudSet():
 * Provide a means to change the baud rate of the running
 * console interface.  If the incoming value is not a valid
 * baud rate, then default to 9600.
 * In the early stages of a new port this can be left empty.
 * Return 0 if successful; else -1.
 */
int
ConsoleBaudSet(int baud)
{
   /* ADD_CODE_HERE */
   return(0);
}

/* target_console_empty():
 * Target-specific portion of flush_console() in chario.c.
 * This function returns 1 if there are no characters waiting to
 * be put out on the UART; else return 0 indicating that the UART
 * is still busy outputting characters from its FIFO.
 * In the early stages of a new port this can simply return 1.
 */
int
target_console_empty(void)
{
  if (( IORD_UART_STATUS(UART1_BASE) & ( UART_STATUS_TMT_MSK | UART_STATUS_TRDY_MSK )) > 0)
  {
      return(0);
  }

  return(1);
}

/* target_putchar():
 * When buffer has space available, load the incoming character
 * into the UART.
 */
int
target_putchar(char c)
{
  unsigned int iTemp;
  
  /* Wait for transmit ready bit */
  while(1)
  {
    iTemp  =  IORD_UART_STATUS(UART1_BASE);
    iTemp &=  UART_STATUS_TMT_MSK;

    if ( iTemp != 0)
    {
      break;
    }  
  }

  IOWR_UART_TXDATA( UART1_BASE, c);

  return((int)c);
}

/* target_gotachar():
 * Return 0 if no char is avaialable at UART rcv fifo; else 1.
 * Do NOT pull character out of fifo, just return status.
 */
int
target_gotachar(void)
{
  unsigned int iTemp;

  // when a character is in the rxd register, RRDY == 1
  iTemp  = IORD_UART_STATUS(UART1_BASE);
  iTemp &= UART_STATUS_RRDY_MSK;

  if ( iTemp != 0)
  {
    return(1);
  }

  return(0);
}

/* target_getchar():
 * Assume there is a character in the UART's input buffer and just
 * pull it out and return it.
 */
int
target_getchar(void)
{
    char    c;

    c = IORD_UART_RXDATA( UART1_BASE);

    return((int)c);
}

/* intsoff():
 * Disable all system interrupts here and return a value that can
 * be used by intsrestore() (later) to restore the interrupt state.
 */
ulong
intsoff(void)
{
   ulong status = 0;
   
   NIOS2_READ_IENABLE( status);
   NIOS2_WRITE_IENABLE(     0);   
   
   return(status);
}

/* intsrestore():
 * Re-establish system interrupts here by using the status value
 * that was returned by an earlier call to intsoff().
 */
void
intsrestore(ulong status)
{
  NIOS2_WRITE_IENABLE( status);
}

/* cacheInitForTarget():
 * Establish target specific function pointers and
 * enable i-cache...
 * Refer to $core/cache.c for a description of the function pointers.
 */
void
cacheInitForTarget(void)
{
  // sorry....nios-ii does not allow the cache to be disabled in software
}

/* target_reset():
 * The default (absolute minimum) action to be taken by this function
 * is to call monrestart(INITIALIZE).  It would be better if there was
 * some target-specific function that would really cause the target
 * to reset...
 */
void
target_reset(void)
{
   monrestart(INITIALIZE);
}

/* If any CPU IO wasn't initialized in reset.S, do it here...
 * This just provides a "C-level" IO init opportunity.
 */
 
// linker define values 
extern void __ram_exc_start;
extern void __rom_exc_start;
extern void __ram_exc_end;

void
initCPUio(void)
{
  // first - copy exception handler to ram vector
   char *p = (char *)&__ram_exc_start;
   char *q = (char *)&__ram_exc_start;
   
   // skip if built for ram
   if ( p != q)
   {
      // copy
      while( p != (char *)&__ram_exc_end)
      {
         *p++ = *q++;
      }
   }
}
