//==========================================================================
//
// pxa250_gpio.c
//
//
// Author(s):    Michael Kelly, Cogent Computer Systems, Inc.
// Contributors: 
// Date:         02/05/2002
// Description:  This file contains code to intialize the PXA250 GPIO
//               section as well as set, clear and test functions
//               for manipulating the GPIO bits
//
//--------------------------------------------------------------------------

#include "config.h"
#include "cpuio.h"
#include "stddefs.h"
#include "genlib.h"
#include "pxa250.h"

// Function Prototypes
void GPIO_init(void);
int GPIO_set(int);
int GPIO_clr(int);
int GPIO_tst(int);
int GPIO_in(int);
int GPIO_out(int);

//--------------------------------------------------------------------------
//
// GPIO assignments for the CSB226 are as follows.
// Active indicates what level is considered active for that
// signal, if any.  A/F indicates alternate function mode (0 = none).
//
// GPIO		Direction	Active	A/F	Description
// 0-1		Input		Low		0	User Switch 0-1
// 2		Input		N/A		0	I2C SCL
// 3-4		Output		Low		0	User LED 0-1
// 5		Input		N/A		0	I2C SDA
// 6		Output		High	0	Unused
// 7		Output		N/A		0	USB Speed, 0 = Low, 1 = High
// 8		Input		N/A		0	Compact FLASH Card Detect Pin 1
// 9		Input		N/A		0	Compact FLASH Card Detect Pin 2
// 10		Input		High	0	Compact FLASH Card Ready Signal
// 11		Output		Low		0	Compact FLASH Power Switch Enable
// 12		Input		Low		0	Compact FLASH Power Switch Error
// 13		Output		High	0	Compact FLASH and CS8900a Reset
// 14		Input		High	0	CS8900a Interrupt
// 15		Output		Low		2	Chip Select 1 - PS/2 Controller
// 16		Input		Low		0	ADS7846E Touch Controller Interrupt
// 17		Output		High	0	LCD Backlight Enable
// 18		Input		High	1	External Bus Ready (used by CS8900a)
// 19-20	Unused
// 21		Input		High	0	PS/2 Keyboard Port Interrupt
// 22		Input		High	0	PS/2 Mouse Port Interrupt
// 23		Output		N/A		1	ADS7846E Touch Controller SCLK
// 24		Output		N/A		1	ADS7846E Touch Controller SFRM
// 25		Output		N/A		1	ADS7846E Touch Controller STXD
// 26		Input		N/A		2	ADS7846E Touch Controller SRXD
// 27		Output		High	0	Passive LCD Display On
// 28		Input		N/A		1	AC97 Bit Clock from Codec
// 29		Input		N/A		1	AC97 Serial Data from Codec
// 30		Output		N/A		2	AC97 Serial Data to Codec
// 31		Output		N/A		2	AC97 SYNC to Codec
// 32		Output		N/A		0	IrDA Transceiver Mode, 1 = High Speed
// 33		Output		Low		2	Chip Select 5 - Unused
// 34		Input		N/A		1	Fast UART RXD
// 35		Input		N/A		1	Fast UART CTS
// 36		Input		N/A		1	Fast UART DCD
// 37		Input		N/A		1	Fast UART DSR
// 38		Input		N/A		1	Fast UART RI
// 39		Output		N/A		2	Fast UART TXD
// 40		Output		N/A		2	Fast UART DTR
// 41		Output		N/A		2	Fast UART RTS
// 42		Input		N/A		1	Bluetooth UART RXD
// 43		Output		N/A		2	Bluetooth UART TXD
// 44		Input		N/A		1	Bluetooth UART CTS
// 45		Output		N/A		2	Bluetooth UART RTS
// 46		Input		N/A		1	IrDA Transceiver RXD
// 47		Output		N/A		1	IrDA Transceiver TXD
// 48		Output		Low		2	Compact FLASH OE
// 49		Output		Low		2	Compact FLASH WE
// 50		Output		Low		2	Compact FLASH IOR
// 51		Output		Low		2	Compact FLASH IOW
// 52		Output		Low		2	Compact FLASH CE1
// 53		Output		Low		2	Compact FLASH CE2
// 54		Output		Low		2	Compact FLASH Socket Select, 0 = Socket 0
// 55		Output		Low		2	Compact FLASH REG
// 56		Input		Low		1	Compact FLASH WAIT
// 57		Input		Low		1	Compact FLASH IOIS16
// 58-73	Output		N/A		2	LCD Data 0-15
// 74		Output		N/A		2	LCD FCLK (VSYNC)
// 75		Output		N/A		2	LCD LCLK (HSYNC)
// 76		Output		N/A		2	LCD PCLK
// 77		Output		Low		2	LCD ACBIAS, Active LCD Enable
// 78		Output		Low		2	Chip Select 2 - CS8900a
// 79		Output		Low		2	Chip Select 3 - Unused
// 80		Output		Low		2	Chip Select 4 - Unused

//--------------------------------------------------------
// GPIO_init()
//
// This function sets the startup state for the PXA250
// as used on the CSB226
//
void
GPIO_init()
{

	// First set the output values to a safe/disabled state before we change any GPIO's to outputs
	// start by setting all of them high which is the safest for most signals
	GPIO_REG(GPIO_GPSR0) = 0xffffffff;
	GPIO_REG(GPIO_GPSR1) = 0xffffffff;
	GPIO_REG(GPIO_GPSR2) = 0xffffffff;

	// Now clear any high true outputs that need to start low
	GPIO_REG(GPIO_GPCR0) = GPIO_7		// USB Speed = High
						 | GPIO_13		// PRST = deasserted
						 | GPIO_17		// BKL_ON
						 | GPIO_27;		// Passive LCD Enable = Off

	// Next, set the correct direction for our functions.  A one means output.
	GPIO_REG(GPIO_GPDR0) = GPIO_3 		// User LED 0
  						 | GPIO_4 		// User LED 1
  						 | GPIO_6 		// LCD Backlight Enable
  						 | GPIO_7 		// USB Speed, 0 = Low, 1 = High
  						 | GPIO_11		// Compact FLASH Power Switch Enable
  						 | GPIO_13		// Compact FLASH and CS8900a Reset
  						 | GPIO_15		// Chip Select 1 - PS/2 Controller
  						 | GPIO_17		// Passive LCD Contrast Voltage (VEE) Enable
  						 | GPIO_23		// ADS7846E Touch Controller SCLK
  						 | GPIO_24		// ADS7846E Touch Controller SFRM
  						 | GPIO_25		// ADS7846E Touch Controller STXD
  						 | GPIO_27		// Passive LCD Display Enable
  						 | GPIO_30		// AC97 Serial Data to Codec
  						 | GPIO_31;		// AC97 SYNC to Codec

	GPIO_REG(GPIO_GPDR1) = GPIO_32		// IrDA Transceiver Mode, 1 = High Speed
  						 | GPIO_33		// Chip Select 5 - Unused
  						 | GPIO_39		// Fast UART TXD
  						 | GPIO_40		// Fast UART DTR
  						 | GPIO_41		// Fast UART RTS
  						 | GPIO_43		// Bluetooth UART TXD
  						 | GPIO_45		// Bluetooth UART RTS
  						 | GPIO_47		// IrDA Transceiver TXD
  						 | GPIO_48		// Compact FLASH OE
  						 | GPIO_49		// Compact FLASH WE
  						 | GPIO_50		// Compact FLASH IOR
  						 | GPIO_51		// Compact FLASH IOW
  						 | GPIO_52		// Compact FLASH CE1
  						 | GPIO_53		// Compact FLASH CE2
  						 | GPIO_54		// Compact FLASH Socket Select, 0 = Socket 0
  						 | GPIO_55		// Compact FLASH REG
  						 | GPIO_58		// LCD Data 0
  						 | GPIO_59		// LCD Data 1
  						 | GPIO_60		// LCD Data 2
  						 | GPIO_61		// LCD Data 3
  						 | GPIO_62		// LCD Data 4
  						 | GPIO_63;		// LCD Data 5

	GPIO_REG(GPIO_GPDR2) = GPIO_64		// LCD Data 6
  						 | GPIO_65		// LCD Data 7
  						 | GPIO_66		// LCD Data 8
  						 | GPIO_67		// LCD Data 9
  						 | GPIO_68		// LCD Data 10
  						 | GPIO_69		// LCD Data 11
  						 | GPIO_70		// LCD Data 12
  						 | GPIO_71		// LCD Data 13
  						 | GPIO_72		// LCD Data 14
  						 | GPIO_73		// LCD Data 15
  						 | GPIO_74		// LCD FCLK (VSYNC)
  						 | GPIO_75		// LCD LCLK (HSYNC)
  						 | GPIO_76		// LCD PCLK
  						 | GPIO_77		// LCD ACBIAS, Active LCD Enable
  						 | GPIO_78		// Chip Select 2 - CS8900a
  						 | GPIO_79		// Chip Select 3 - Unused
  						 | GPIO_80;		// Chip Select 4 - Unused

	// Finally, set the alternate function registers to the correct state
	GPIO_REG(GPIO_GAFR0L) = GPIO_15_AF_CS1;				// AF2	Chip Select 1 - PS/2 Controller

	GPIO_REG(GPIO_GAFR0U) = GPIO_18_AF_RDY				// AF1	External Bus Ready (used by CS8900a)
				 		  | GPIO_23_AF_SSP_CLK			// AF1	ADS7846E Touch Controller SCLK
				 		  | GPIO_24_AF_SSP_FRM			// AF1	ADS7846E Touch Controller SFRM
				 		  | GPIO_25_AF_SSP_TXD			// AF1	ADS7846E Touch Controller STXD
				 		  | GPIO_26_AF_SSP_RXD			// AF2	ADS7846E Touch Controller SRXD
				 		  | GPIO_28_AF_AC97_BCLK_IN 	// AF1	AC97 Bit Clock from Codec
				 		  | GPIO_29_AF_AC97_SDIN0		// AF1	AC97 Serial Data from Codec
				 		  | GPIO_30_AF_AC97_SDOUT		// AF2	AC97 Serial Data to Codec
				 		  | GPIO_31_AF_AC97_SYNC;		// AF2	AC97 SYNC to Codec

	GPIO_REG(GPIO_GAFR1L) = GPIO_33_AF_CS5				// AF2	Chip Select 5 - Unused
				  		  | GPIO_34_AF_FF_RXD			// AF1	Fast UART RXD
				  		  | GPIO_35_AF_FF_CTS			// AF1	Fast UART CTS
				  		  | GPIO_36_AF_FF_DCD			// AF1	Fast UART DCD
				  		  | GPIO_37_AF_FF_DSR			// AF1	Fast UART DSR
				  		  | GPIO_38_AF_FF_RI			// AF1	Fast UART RI
				  		  | GPIO_39_AF_FF_TXD			// AF2	Fast UART TXD
				  		  | GPIO_40_AF_FF_DTR			// AF2	Fast UART DTR
				  		  | GPIO_41_AF_FF_RTS			// AF2	Fast UART RTS
				  		  | GPIO_42_AF_BT_RXD			// AF1	Bluetooth UART RXD
				  		  | GPIO_43_AF_BT_TXD			// AF2	Bluetooth UART TXD
				  		  | GPIO_44_AF_BT_CTS			// AF1	Bluetooth UART CTS
				  		  | GPIO_45_AF_BT_RTS			// AF2	Bluetooth UART RTS
				  		  | GPIO_46_AF_IR_RXD			// AF1	IrDA Transceiver RXD
				  		  | GPIO_47_AF_IR_TXD;			// AF1	IrDA Transceiver TXD

	GPIO_REG(GPIO_GAFR1U) = GPIO_48_AF_POE				// AF2	Compact FLASH OE
				  		  | GPIO_49_AF_PWE				// AF2	Compact FLASH WE
				  		  | GPIO_50_AF_PIOR				// AF2	Compact FLASH IOR
				  		  | GPIO_51_AF_PIOW				// AF2	Compact FLASH IOW
				  		  | GPIO_52_AF_PCE1				// AF2	Compact FLASH CE1
				  		  | GPIO_53_AF_PCE2				// AF2	Compact FLASH CE2
				  		  | GPIO_54_AF_PSKTSEL			// AF2	Compact FLASH Socket Select, 0 = Socket 0
				  		  | GPIO_55_AF_PREG				// AF2	Compact FLASH REG
				  		  | GPIO_56_AF_PWAIT			// AF1	Compact FLASH WAIT
				  		  | GPIO_57_AF_IOIS16			// AF1	Compact FLASH IOIS16
				  		  | GPIO_58_AF_LDD0				// AF2	LCD Data 0
				  		  | GPIO_59_AF_LDD1				// AF2	LCD Data 1
				  		  | GPIO_60_AF_LDD2				// AF2	LCD Data 2
				  		  | GPIO_61_AF_LDD3				// AF2	LCD Data 3
				  		  | GPIO_62_AF_LDD4				// AF2	LCD Data 4
				  		  | GPIO_63_AF_LDD5;			// AF2	LCD Data 5

	GPIO_REG(GPIO_GAFR2L) = GPIO_64_AF_LDD6				// AF2	LCD Data 6
						  | GPIO_65_AF_LDD7				// AF2	LCD Data 7
						  | GPIO_66_AF_LDD8				// AF2	LCD Data 8
						  | GPIO_67_AF_LDD9				// AF2	LCD Data 9
						  | GPIO_68_AF_LDD10			// AF2	LCD Data 10
						  | GPIO_69_AF_LDD11			// AF2	LCD Data 11
						  | GPIO_70_AF_LDD12			// AF2	LCD Data 12
						  | GPIO_71_AF_LDD13			// AF2	LCD Data 13
						  | GPIO_72_AF_LDD14			// AF2	LCD Data 14
						  | GPIO_73_AF_LDD15			// AF2	LCD Data 15
						  | GPIO_74_AF_LCD_FCLK			// AF2	LCD FCLK (VSYNC)
						  | GPIO_75_AF_LCD_LCLK			// AF2	LCD LCLK (HSYNC)
						  | GPIO_76_AF_LCD_PCLK			// AF2	LCD PCLK
						  | GPIO_77_AF_LCD_BIAS			// AF2	LCD ACBIAS (LCD ON)
						  | GPIO_78_AF_CS2				// AF2	Chip Select 2 - CS8900a
						  | GPIO_79_AF_CS3;				// AF2	Chip Select 3 - Unused

	GPIO_REG(GPIO_GAFR2U) = GPIO_80_AF_CS4;				// AF2	Chip Select 4 - Unused

	// finally, clear the RDH bit (by writing 1 to it) in the PSSR to enable GPIO inputs
	PWR_REG(PWR_PSSR) = 0x20;
}

//--------------------------------------------------------
// GPIO_set()
//
// This function sets the desired bit passed in.  It first 
// tests to see if the bit is an actual output.  If not,
// it returns a -1, otherwise it returns 0.  NOTE: We do not
// test to see if setting the bit would screw up any alternate
// functions.  Use this function with caution!
//

int GPIO_set(int gpio_bit)
{
	// quick sanity test
	if (gpio_bit > 80) return -1;

	// First find out which group the bit belongs to
	if (gpio_bit < 32)
	{
		// Get the direction and see if it's an output
		if (!(GPIO_REG(GPIO_GPDR0) & (1 << gpio_bit)))
			return -1;

		// if yes, then set it
		GPIO_REG(GPIO_GPSR0) = (1 << gpio_bit);
	}
	else if (gpio_bit < 64)
	{
		// Get the direction and see if it's an output
		if (!(GPIO_REG(GPIO_GPDR1) & (1 << (gpio_bit - 32))))
			return -1;

		// if yes, then set it
		GPIO_REG(GPIO_GPSR1) = (1 << (gpio_bit - 32));
	}
	else
	{
		// Get the direction and see if it's an output
		if (!(GPIO_REG(GPIO_GPDR2) & (1 << (gpio_bit - 64))))
			return -1;

		// if yes, then set it
		GPIO_REG(GPIO_GPSR2) = (1 << (gpio_bit - 64));
	}
	return 0;
}

//--------------------------------------------------------
// GPIO_clr()
//
// This function clears the desired bit passed in.  It first 
// tests to see if the bit is an actual output.  If not,
// it returns a -1, otherwise it returns 0.  NOTE: We do not
// test to see if clearing the bit would screw up any alternate
// functions.  Use this function with caution!
//

int GPIO_clr(int gpio_bit)
{
	// quick sanity test
	if (gpio_bit > 80) return -1;

	// First find out which group the bit belongs to
	if (gpio_bit < 32)
	{
		// Get the direction and see if it's an output
		if (!(GPIO_REG(GPIO_GPDR0) & (1 << gpio_bit)))
			return -1;

		// if yes, then set it
		GPIO_REG(GPIO_GPCR0) = (1 << gpio_bit);
	}
	else if (gpio_bit < 64)
	{
		// Get the direction and see if it's an output
		if (!(GPIO_REG(GPIO_GPDR1) & (1 << (gpio_bit - 32))))
			return -1;

		// if yes, then clear it
		GPIO_REG(GPIO_GPCR1) = (1 << (gpio_bit - 32));
	}
	else
	{
		// Get the direction and see if it's an output
		if (!(GPIO_REG(GPIO_GPDR2) & (1 << (gpio_bit - 64))))
			return -1;

		// if yes, then clear it
		GPIO_REG(GPIO_GPCR2) = (1 << (gpio_bit - 64));
	}
	return 0;
}
//--------------------------------------------------------
// GPIO_tst()
//
// This function returns the state of desired bit passed in.
// It does not test to see if it's an input or output and thus
// can be used to verify if an output set/clr has taken place
// as well as for testing an input state.
//

int GPIO_tst(int gpio_bit)
{
	// First find out which group the bit belongs to
	if (gpio_bit < 32)
	{
		// Get the state
		if (GPIO_REG(GPIO_GPLR0) & (1 << gpio_bit)) return 1;
	}
	else if (gpio_bit < 64)
	{
		// Get the state
		if (GPIO_REG(GPIO_GPLR1) & (1 << (gpio_bit - 32))) return 1;
	}
	else
	{
		// Get the state
		if (GPIO_REG(GPIO_GPLR2) & (1 << (gpio_bit - 64))) return 1;
	}
	return 0; // bit was not set
}

//--------------------------------------------------------
// GPIO_in()
//
// This function changes the direction of the desired bit 
// to input.  NOTE: We do not test to see if changing the
// direction of the bit would screw up anything.  Use this
// function with caution!
int GPIO_in(int gpio_bit)
{
	// quick sanity test
	if (gpio_bit > 80) return -1;

	// now find out which group the bit belongs to
	if (gpio_bit < 32)
	{
		GPIO_REG(GPIO_GPDR0) &= ~(1 << gpio_bit);
	}
	else if (gpio_bit < 64)
	{
		GPIO_REG(GPIO_GPDR1) &= ~(1 << (gpio_bit - 32));
	}
	else
	{
		GPIO_REG(GPIO_GPDR2) &= ~(1 << (gpio_bit - 64));
	}
	return 0;
}

//--------------------------------------------------------
// GPIO_out()
//
// This function changes the direction of the desired bit 
// to output.  NOTE: We do not test to see if changing the
// direction of the bit would screw up anything.  Use this
// function with caution!
int GPIO_out(int gpio_bit)
{
	// quick sanity test
	if (gpio_bit > 80) return -1;

	// now find out which group the bit belongs to
	if (gpio_bit < 32)
	{
		GPIO_REG(GPIO_GPDR0) |= (1 << gpio_bit);
	}
	else if (gpio_bit < 64)
	{
		GPIO_REG(GPIO_GPDR1) |= (1 << (gpio_bit - 32));
	}
	else
	{
		GPIO_REG(GPIO_GPDR2) |= (1 << (gpio_bit - 64));
	}
	return 0;
}

