//=============================================================================
//
//      cpuio.c
//
//      CPU/Board Specific IO
//
// Author(s):    Michael Kelly, Cogent Computer Systems, Inc.
// Date:         05-26-2002
// Description:  This file contains the IO functions required by Micro Monitor
//				 that are unique to each CPU/Board combination
//
//
//=============================================================================

#include "config.h"
#include "cpuio.h"
#include "genlib.h"
#include "ether.h"
#include "stddefs.h"
#include "warmstart.h"
#include "ml67q5003.h"
#include "cpuio.h"

extern int cpu_type_4k;		// modified in hw_init2.c

// Delay for some usecs. - Not accurate, assumes ROM mode
// and no Cache
void udelay(int delay)
{
  int i;
  for ( i = LOOPS_PER_USEC * delay; i ; i--)
	;
}

/* InitUART():
 * This is the top-level UART initialization for the monitor. 
 */
void
InitUART(int baud)
{
	// The ML67Q5003 derives it's baud clock from the input clock.
	// Baud = 256 - (CCLK/(Baud * 16))
	SIO_REG(SIO_BCN) = 0;	// stop the buad rate counter first

	if (cpu_type_4k){
		SIO_REG(SIO_BT) = ((256 * (16 * baud) - CCLK_4K) / (16 * baud));
	}
	else {
		SIO_REG(SIO_BT) = ((256 * (16 * baud) - CCLK_5K) / (16 * baud));
	}

	SIO_REG(SIO_BCN) = SIO_BCN_EN;	// start

	// set the port to 8-bits, 1 stop
	SIO_REG(SIO_CON) = SIO_CON_LN8 | SIO_CON_STP1;

	// write a carriage return to force the
	// transmit ready bit to go high after it is completed
	SIO_REG(SIO_BUF) = 0x0a;

}

int
ConsoleBaudSet(int baud)
{
	/* This uart doesn't like 9600 for some reason!!??
 	 */
	switch(baud) {
		case 19200:
		case 38400:
		case 57600:
		case 115200:
			InitUART(baud);
			return(0);
		default:
			return(-1);
	}
}

int
devInit(int baud)
{

	// Initialize the UART
    InitUART(baud);

	// Write 4 or 5 to the LED
	// depending upon what we are
	LED_DP_OFF();

	ssled(0);

	// release the SMC from reset
	GPIOC_REG(GPIO_PO) &= ~0x0080;

	return 0;
}

/* target_console_empty():
 * Target-specific portion of flush_console() in chario.c.
 * This function returns 1 if there are no characters waiting to
 * be put out on the UART; else return 0 indicating that the UART
 * is still busy outputting characters from its FIFO.
 */
int
target_console_empty(void)
{
	if (SIO_REG(SIO_STA) & SIO_STA_TXRDY)
		return(1);
	return(0);
}

/* target_putchar():
 */
int
target_putchar(char c)
{
    /* Wait for transmit ready bit */
    while(1) {
        if (SIO_REG(SIO_STA) & SIO_STA_TXRDY)
            break;
    }

	// clear the tx ready bit first
	SIO_REG(SIO_STA) = SIO_STA_TXRDY;

	// now write the character
	SIO_REG(SIO_BUF) = c;

    return((int)c);
}

/* target_gotachar():
 * Return 0 if no char is avaialable at UART rcv fifo; else 1.
 * Do NOT pull character out of fifo, just return status. 
 */
int
target_gotachar()
{
    if (!(SIO_REG(SIO_STA) & SIO_STA_RXRDY))
        return(0);
	else
	    return(1);
}

/* target_getchar():
 */
int 
target_getchar(void)
{
    char    c;

    /* Retrieve character
	 */
    c = SIO_REG(SIO_BUF);

	/* Then clear the rx ready bit
	 */
    SIO_REG(SIO_STA) = SIO_STA_RXRDY;
    return((int)c);
}

void
target_reset(void)
{
	warmstart(INITIALIZE);
}

void
ssled(int value)
{
	switch(value) {
		case 0x0:
			LED_0();
			break;
		case 0x1:
			LED_1();
			break;
		case 0x2:
			LED_2();
			break;
		case 0x3:
			LED_3();
			break;
		case 0x4:
			LED_4();
			break;
		case 0x5:
			LED_5();
			break;
		case 0x6:
			LED_6();
			break;
		case 0x7:
			LED_7();
			break;
		case 0x8:
			LED_8();
			break;
		case 0x9:
			LED_9();
			break;
		case 0xA:
			LED_A();
			break;
		case 0xB:
			LED_B();
			break;
		case 0xC:
			LED_C();
			break;
		case 0xD:
			LED_D();
			break;
		case 0xE:
			LED_E();
			break;
		case 0xF:
			LED_F();
			break;
		default:
			break;
	}
}

void
initCPUio()
{
	cpu_type_4k = 0;	/* CPU is 5K */
}

extern ulong getpsr(void);
extern void putpsr(ulong);

void
intsrestore(ulong psr)
{
    putpsr(psr);
}

/*
 * Read the program status register (CPSR)
 * and set the FIQ and IRQ bits.
 */
ulong
intsoff(void)
{
    ulong  psr;

    psr = getpsr();

    /*
     * Set bit 6, bit 7 to disable interrupts.
     */
    putpsr(psr | 0x000000c0);
    return(psr);
}

/* cacheInitForTarget():
    Enable instruction cache only...
*/
void
cacheInitForTarget()
{
//    asm("   MRC p15, 0, r0, c1, c0, 0");
//    asm("   ORR r0, r0, #0x1000");  /* bit 12 is ICACHE enable*/
//    asm("   MCR p15, 0, r0, c1, c0, 0");

    /* Flush instruction cache */
//    asm("   MCR p15, 0, r0, c7, c5, 0");
}
