/* fat.h

    Author: David de la Cruz, Technology Rescue.
    email:  ddlc@techresc.net

	This include file defines FAT Filesystem structures and related defs.

	Modification History :

	Version 2.1:
		29.04.2004 : new.
*/

#ifndef __fat_h_
#define	__fat_h_

/* Partition entry. */
typedef struct dos_partition
{
	uchar boot_ind;			/* 0x80 = boot device - 0x00 = non-bootable */
	uchar head;				/* starting head */
	uchar sector;			/* starting sector */
	uchar cyl;				/* starting cylinder */
	uchar sys_ind;			/* What partition type */
	uchar end_head;			/* end head	*/
	uchar end_sector;		/* end sector */
	uchar end_cyl;			/* end cylinder	*/
	ulong start;			/* starting sector counting from 0 */
	ulong size;				/* nr of sectors in partition */
} dos_partition;

/* Master boot record structure is used to contain sector 0. */
typedef struct master_boot_record
{
	uchar reserved[446];							/* This area is used to store Consistency Check Routine if any. */
	dos_partition Part[4] __attribute__ ((packed));	/* 4 partition entries. */
	ushort signature __attribute__ ((packed));		/* should be 0xAA55 */
} master_boot_record;


typedef struct boot_sector_fat32
{
	ulong	fat32_length;							/* Sectors/FAT */
	ushort	flags;									/* Bit 8: fat mirroring, low 4: active fat */
	ushort	version;								/* Filesystem version */
	ulong	root_cluster;							/* First cluster of the root directory */
	ushort	info_sector;							/* Sector number of FSINFO structure */
	ushort	backup_boot;							/* Backup boot sector */
	uchar	reserved2[12];							/* Unused */
	uchar drive_number __attribute__ ((packed));	/* drive number */
	uchar reserved __attribute__ ((packed));		/* Unused */
	uchar ext_boot_sign __attribute__ ((packed));	/* 0x29 if fields below exist (DOS 3.3+) */
	ulong volume_id __attribute__ ((packed));		/* Volume ID number */
	char volume_label[11] __attribute__ ((packed));	/* Volume label */
	char fs_type[8] __attribute__ ((packed));		/* Typically FAT12, FAT16, or FAT32 */
	uchar unused[420] __attribute__ ((packed));		/* Load Program Code if any. */
	ushort signature __attribute__ ((packed));		/* should be 0xAA55 */
} boot_sector_fat32;

typedef struct boot_sector_fat16
{
	uchar drive_number;								/* drive number */
	uchar reserved __attribute__ ((packed));		/* Unused */
	uchar ext_boot_sign __attribute__ ((packed));	/* 0x29 if fields below exist (DOS 3.3+) */
	ulong volume_id __attribute__ ((packed));		/* Volume ID number */
	char volume_label[11] __attribute__ ((packed));	/* Volume label */
	char fs_type[8] __attribute__ ((packed));		/* Typically FAT12, FAT16, or FAT32 */
	uchar unused[448] __attribute__ ((packed));		/* Load Program Code if any. */
	ushort signature __attribute__ ((packed));		/* should be 0xAA55 */
} boot_sector_fat16;

typedef union boot_sector_fatxx
{
        boot_sector_fat16 fat16;
        boot_sector_fat32 fat32;
} boot_sector_fatxx;

/* boot_sector is used to contain either FAT16 or FAT32 boot sector. */
typedef struct boot_sector
{
	/* Common data structure */
	uchar	bootstrap[3];							/* Bootstrap code */
	char	system_id[8] __attribute__ ((packed));	/* OEM Name */
	ushort	sector_size __attribute__ ((packed));	/* Bytes/sector */
	uchar	cluster_size __attribute__ ((packed));	/* Sectors/cluster */
	ushort	reserved __attribute__ ((packed));		/* Number of reserved sectors */
	uchar	fats __attribute__ ((packed));			/* Number of FATs */
	ushort	dir_entries __attribute__ ((packed));	/* Number of root directory entries */
	ushort	sectors __attribute__ ((packed));		/* Number of sectors */
	uchar	media __attribute__ ((packed));			/* Media code */
	ushort	fat_length __attribute__ ((packed));	/* Sectors/FAT */
	ushort	secs_track __attribute__ ((packed));	/* Sectors/track */
	ushort	heads __attribute__ ((packed));			/* Number of heads */
	ulong	hidden __attribute__ ((packed));		/* Number of hidden sectors */
	ulong	total_sect __attribute__ ((packed));	/* Number of sectors (if sectors == 0) */
	boot_sector_fatxx ext __attribute__ ((packed));	/* Extended boot sector info : either FAT16 or FAT32 */
} boot_sector;


/* dir_entry is used to find directory entries in FAT sectors. */
typedef struct dir_entry
{
	char	name[FILENAME_LEN+FILENAME_EXT] __attribute__ ((packed)); /* name + extension of file name. */
	uchar	attr __attribute__ ((packed));			/* Attribute bits */
	uchar	reserved __attribute__ ((packed));		/* Reserved for use by Windows NT. */
	uchar	ctime_ms __attribute__ ((packed));		/* Millisecond stamp at file creation time. */
	ushort	ctime __attribute__ ((packed));			/* Time file was created. */
	ushort	cdate __attribute__ ((packed));			/* Date file was created. */
	ushort	adate __attribute__ ((packed));			/* Last access date. */
	ushort	starthi __attribute__ ((packed));		/* High 16 bits of cluster in FAT32 */
	ushort	time __attribute__ ((packed));			/* Time of last write. */
	ushort	date __attribute__ ((packed));			/* Date of last write. */
	ushort	start __attribute__ ((packed));			/* First cluster. */
	ulong	size __attribute__ ((packed));			/* File size in bytes. */
} dir_entry;


/* Following structure contains file system data. */
typedef struct file_system
{
	/* Define offsets related to partition start. */
	ulong	boot_sector_off;/* Boot sector offset related to partition start. */
	ulong	fat_off;		/* FAT sector offset related to partition start. */
	ulong	fat_save_off;	/* Saved FAT sector. */
	ulong	root_dir_off;	/* Root directory offset. */
	ulong	data_off;		/* First Cluster offset. */

	/* Following fields keep file system informations computed according to Microsoft Documentation :
	FAT: General Overview of On-Disk Format; Version 1.02, May 5, 1999. */
	ulong	root_dir;		/* # sectors in root directory. */
	ulong	fat_size;		/* # sector per FAT. */
	ulong	cluster_size;	/* cluster size in bytes. */
	ulong	tot_sec;		/* # total sectors */
	ulong	data_sec;		/* # data sectors. */
	ulong	count_of_clus;	/* Count of clusters. This field is used for fat type determination. */
	ulong	fat_type;		/* Keep fat type.*/

	boot_sector * BPB;		/* Keep pointer to bootsector infos. */
} file_system;

typedef struct file_infos
{
	char filename[FILENAME_SIZE]; /* 8+3 formated and NULL terminated string. */ 
	ulong dir_entry_addr;	/* Directory entry address. */
	ulong cluster_num;		/* First cluster of cluster chain. */
	ulong file_size;		/* File size in bytes. */
	ulong current_fat_entry;/* Address of fat entry for the current cluster. */
	ulong current_cluster;	/* Current cluster address. */
#if 0
} file_infos __attribute__ ((packed));
#else
} file_infos;
#endif


#ifndef NULL
#define NULL	((void *) 0)
#endif

/* File attributes defs. */
#define ATTR_RO     0x01
#define ATTR_HIDDEN 0x02
#define ATTR_SYS    0x04
#define ATTR_VOLUME 0x08
#define ATTR_DIR    0x10
#define ATTR_ARCH   0x20
#define ATTR_VFAT   (ATTR_RO | ATTR_HIDDEN | ATTR_SYS | ATTR_VOLUME)

/* Directory defs. */
#define	DIRENT_FREE	0x5E
#define	DIRENT_LAST	0x00

/* Partition types and flags. */
#define PDOS12		1	/* DOS 12-bit FAT < 16 MB. */
#define PDOS16		4	/* DOS 16-bit FAT < 32 MB. */
#define PDOSEXT		5	/* Extended DOS. */
#define PDOS32		6	/* DOS 16-bit FAT >= 32 MB. */
#define BOOTABLE	0x80
#define NBOOTABLE	0x00

/* Sector size. */
#define FS_BLOCK_SIZE 512

/* Last cluster chain id. */
#define PDOS16_EOC	0xFFFF
#define PDOS32_EOC	0x0FFFFFFF


/* Function prototypes. */

/* init_media function.

	Entry :
		file_system_ops * fsops : filled in with media functions.

	Exit :
		return 0 if successfull, else error code.

	Function :
		It initialises list_of_media.
		This function is called by media module such as mmc.

*/
int init_media(file_system_ops * fsops);

/* init_sector0 function.

	Entry :
		master_boot_record * MBR : master boot record initialised with 0s.
		ulong * sector0 : uninitialised buffer.

	Exit :
		master_boot_record * MBR : master boot record initialised with first sector.
		ulong * sector0 : buffer which contains sector 0 datas.
		return 0 if successfull, else error code.

	Notes :
		MBR and sector0 may contain same data if media is floppy-like.
*/
int init_sector0(master_boot_record * MBR, ulong * sector0);

/* init_boot_sector function.

	Entry :
		boot_sector * BPB : boot sector structure initialised with 0s.
		ulong * sector0 : buffer which contains sector 0 datas.

	Exit :
		boot_sector * BPB : boot sector structure filled in with sector0 area.
		return 0 if successfull, else error code.
*/
int init_boot_sector(boot_sector * BPB, ulong * sector0);

/* init_file_system function.

	Entry :
		file_system * fs : file system structure initialised with 0s.
		boot_sector * BPB : boot sector structure initialised by init_boot_sector function.

	Exit :
		file_system * fs : file system structure filled in with boot sector infos.
		return 0 if successfull, else error code.
*/
int init_file_system(file_system * fs, boot_sector * BPB);


/* print_file_system function.

	Entry :
		file_system * fs : file system structure to print.

	Exit :
		no return
*/
void print_file_system(file_system * fs);

/* print_mbr function.

	Entry :
		master_boot_record * mbr : file system structure to print.

	Exit :
		no return
*/
void print_mbr(master_boot_record * mbr);

/* print_boot_sector function.

	Entry :
		boot_sector * bs : file system structure to print.

	Exit :
		no return
*/
void print_boot_sector(boot_sector * bs);

/* print_file_infos function.

	Entry :
		file_infos * fi : file infos structure to print.

	Exit :
		no return
*/
void print_file_infos(file_infos * fi);


/* find_filename function.

	Entry :
		const char * filename : filename string to search.
		const file_system * fs : file system structure.

	Exit :
		file_infos structure filled in with directory entry content.
		file_infos->cluster_num = 0 if file not found.
*/
file_infos find_filename(const char * filename, const file_system * fs);

/* read_cluster function.

	Entry :
		const file_infos * fi : file info structure.
		const file_system * fs : file system structure.
		ulong dest : destination address to write to.

	Exit :
		<0 : error in file read.
		>0 : bytes read from file and written to dest address.
*/
int read_cluster(const file_infos * fi, const file_system * fs, ulong dest);

/* find_next_cluster function.

	Entry :
		const file_infos * fi : contains current file infos.
		const file_system * fs : file system structure.

	Exit :
		file_infos structure filled in with fi updated = next cluster infos.
		file_infos->cluster_num = 0 if file not found.
*/
file_infos find_next_cluster(const file_infos * fi, const file_system * fs);

/* init_dir_entry function.

	Entry :
		file_infos * fi : contains current file infos.
		const file_system * fs : file system structure.

	Exit :
		file_infos * fi : contains updated file infos such as file name

*/
void init_dir_entry(file_infos * fi, const file_system * fs);

/* find_free_cluster function.

	Entry :
		const file_infos * fi : contains current file infos.
		const file_system * fs : file system structure.

	Exit :
		file_infos structure filled in with fi updated = next cluster infos.
		file_infos->cluster_num = 0 if file not found.
*/
file_infos find_free_cluster(const file_infos * fi, const file_system * fs);

/* write_cluster function.

	Entry :
		const file_infos * fi : file info structure.
		const file_system * fs : file system structure.
		ulong src : source address to read from.

	Exit :
		<0 : error in file read.
		>0 : bytes read from file and written to dest address.
*/
int write_cluster(const file_infos * fi, const file_system * fs, ulong src);


/* update_fat function.

	Entry :
		file_infos * fi : current file info structure to update with next_fi->next_cluster_num.
		const file_infos * next_fi : next file info structure.
		const file_system * fs : file system structure.

	Exit :
		file_infos * fi : updated file info structure with next cluster number.
		<0 : error.
		0 : update was successfully done.
*/
int update_fat(file_infos * fi, const file_infos * next_fi, const file_system * fs);

/* update_file_system function.

	Entry :
		file_infos * fi : contains file info structure with first cluster number and file size.
		const file_system * fs : file system structure.

	Exit :
		<0 : error.
		>0 : size of file.

*/
int update_file_system(const file_infos * fi, const file_system * fs);


#endif	/* __fat_h_ */
