/* mmc.c:
 *	This file contains the portions of the flash code that are device
 *	independent.  Refer to the appropriate device sub-directory for the
 *	code that is specific to the flash device on the target.
 *
 *	General notice:
 *	This code is part of a boot-monitor package developed as a generic base
 *	platform for embedded system designs.  As such, it is likely to be
 *	distributed to various projects beyond the control of the original
 *	author.  Please notify the author of any enhancements made or bugs found
 *	so that all may benefit from the changes.  In addition, notification back
 *	to the author will allow the new user to pick up changes that may have
 *	been made by other users after this version of the code was distributed.
 *
 *	Note1: the majority of this code was edited with 4-space tabs.
 *	Note2: as more and more contributions are accepted, the term "author"
 *		   is becoming a mis-representation of credit.
 *
 *	Original author:	David de la Cruz
 *	Email:				ddlc@techresc.net
 *
 *	Technology Rescue History:
 *	Version 2.1 :
 *		26.04.2004 : new module.
 *		28.04.2004 : added SDCardType flag and fixed initialisation flow.
 *		30.04.2004 : added FAT support.
 *		02.05.2004 : fixed MmcRead and MmcWrite return values.
 *		06.05.2004 : clean-up.
 *
 *	Notes :
 *		Only one card is supported.
 */

#include "config.h"
#if INCLUDE_MMC
#include "cpu.h"
#include "genlib.h"
#include "ctype.h"
#include "stddefs.h"
#include "tfs.h"
#include "cli.h"
#include "mmc.h"
#include "pxa250_mmc.h"
#include "pxa250.h"
#include "errno.h"
#if INCLUDE_FS
#include "fs.h"
#if INCLUDE_FAT
#include "fat.h"
#endif
#endif


static uchar mmc_buf[MMC_BLOCK_SIZE];
static mmc_csd_t mmc_csd;
static mmc_sd_cid_t CID;
static int mmc_ready = 0;
static int SDCardType = 0;			// Assume Card is MultiMediaCard.
static int RCA = MMC_DEFAULT_RCA;	// Default address is MMC_DEFAULT_RCA for MMC card.
									// For SD Card, RCA is read from CSD response.

#if INCLUDE_FS
file_system_ops mmc_fsops;			// This structure is used to register mmc into media list.
#endif

static uchar *
/****************************************************/
mmc_cmd(ushort cmd, ushort argh, ushort argl, ushort cmdat)
/****************************************************/
{
	static uchar resp[20];
	ulong status;
	int words, i;

#ifdef MMC_DEBUG
	printf("mmc_cmd %x %x %x %x\n", cmd, argh, argl, cmdat);
#endif
	MMC_REG(MMC_MMCSTRPCL) = MMC_STRPCL_STOP_CLK;
	while((MMC_REG(MMC_MMCSTAT) & MMC_STAT_CLK_EN) == MMC_STAT_CLK_EN)
	MMC_REG(MMC_MMCARGH)   = argh;
	MMC_REG(MMC_MMCARGL)   = argl;
	MMC_REG(MMC_MMCCMD)    = cmd;
	MMC_REG(MMC_MMCCMDAT)  = cmdat;
	MMC_REG(MMC_MMCIMASK) = ~MMC_I_MASK_END_CMD_RES;
	MMC_REG(MMC_MMCSTRPCL) = MMC_STRPCL_START_CLK;
	while (!(MMC_REG(MMC_MMCSTAT) & MMC_STAT_END_CMD_RES));

	status = MMC_REG(MMC_MMCSTAT);
#ifdef MMC_DEBUG
	printf("MMC status %x\n", status);
#endif
	if (status & MMC_STAT_TIME_OUT_RESPONSE)
	{
		return 0;
	}

	switch (cmdat & 0x3)
	{
		case MMC_CMDAT_R1:
		case MMC_CMDAT_R3:
			words = 3;
			break;

		case MMC_CMDAT_R2:
			words = 8;
			break;

		default:
			return 0;
	}
	for (i = words-1; i >= 0; i--)
	{
		ulong res_fifo = MMC_REG(MMC_MMCRES);
		int offset = i << 1;

		resp[offset] = ((uchar *)&res_fifo)[0];
		resp[offset+1] = ((uchar *)&res_fifo)[1];
	}
#ifdef MMC_DEBUG
	for (i=0; i<words*2; i += 2)
	{
		printf("MMC resp[%d] = %02x\n", i, resp[i]);
		printf("MMC resp[%d] = %02x\n", i+1, resp[i+1]);
	}
#endif
	return resp;
}

int
/****************************************************/
mmc_block_read(uchar *dst, ulong src, ulong len)
/****************************************************/
{
	uchar *resp;
	ushort argh, argl;
	ulong status;

	if (len == 0)
	{
		return 0;
	}

#ifdef MMC_DEBUG
	printf("mmc_block_rd dst %lx src %lx len %d\n", (ulong)dst, src, len);
#endif

	argh = len >> 16;
	argl = len & 0xffff;

	/* set block len */
	resp = mmc_cmd(MMC_CMD_SET_BLOCKLEN, argh, argl, MMC_CMDAT_R1);

	/* send read command */
	argh = src >> 16;
	argl = src & 0xffff;
	MMC_REG(MMC_MMCSTRPCL) = MMC_STRPCL_STOP_CLK;
	MMC_REG(MMC_MMCRDTO) = 0xffff;
	MMC_REG(MMC_MMCNOB) = 1;
	MMC_REG(MMC_MMCBLKLEN) = len;
	resp = mmc_cmd(MMC_CMD_READ_BLOCK, argh, argl,
			MMC_CMDAT_R1|MMC_CMDAT_READ|MMC_CMDAT_BLOCK|MMC_CMDAT_DATA_EN);


	MMC_REG(MMC_MMCIMASK) = ~MMC_I_MASK_RXFIFO_RD_REQ;
	while (len)
	{
		if (MMC_REG(MMC_MMCIREG) & MMC_I_REG_RXFIFO_RD_REQ)
		{
			*dst++ = MMC_REG(MMC_MMCRXFIFO);
			len--;
		}
		status = MMC_REG(MMC_MMCSTAT);
		if (status & MMC_STAT_ERRORS)
		{
			printf("MMC_STAT error %lx\n", status);
			return -1;
		}
	}
	MMC_REG(MMC_MMCIMASK) = ~MMC_I_MASK_DATA_TRAN_DONE;
	while (!(MMC_REG(MMC_MMCIREG) & MMC_I_REG_DATA_TRAN_DONE));
	status = MMC_REG(MMC_MMCSTAT);
	if (status & MMC_STAT_ERRORS)
	{
		printf("MMC_STAT error %lx\n", status);
		return -1;
	}
	return 0;
}

int
/****************************************************/
mmc_block_write(ulong dst, uchar *src, int len)
/****************************************************/
{
	uchar *resp;
	ushort argh, argl;
	ulong status;

#ifdef MMC_DEBUG
	printf("MmcRead dst %lx src %lx len %d\n", dst, (ulong)src, len);
#endif

	if (len == 0)
	{
		return 0;
	}

#ifdef MMC_DEBUG
	printf("mmc_block_wr dst %lx src %lx len %d\n", dst, (ulong)src, len);
#endif

	argh = len >> 16;
	argl = len & 0xffff;

	/* set block len */
	resp = mmc_cmd(MMC_CMD_SET_BLOCKLEN, argh, argl, MMC_CMDAT_R1);

	/* send write command */
	argh = dst >> 16;
	argl = dst & 0xffff;
	MMC_REG(MMC_MMCSTRPCL) = MMC_STRPCL_STOP_CLK;
	MMC_REG(MMC_MMCNOB) = 1;
	MMC_REG(MMC_MMCBLKLEN) = len;
	resp = mmc_cmd(MMC_CMD_WRITE_BLOCK, argh, argl,
			MMC_CMDAT_R1|MMC_CMDAT_WRITE|MMC_CMDAT_BLOCK|MMC_CMDAT_DATA_EN);

	MMC_REG(MMC_MMCIMASK) = ~MMC_I_MASK_TXFIFO_WR_REQ;
	while (len)
	{
		if (MMC_REG(MMC_MMCIREG) & MMC_I_REG_TXFIFO_WR_REQ)
		{
			int i, bytes;
			if(len < 32)	// equiv to  = min(32,len)
			{
				bytes = len;
			}
			else
			{
				bytes = 32;
			}

			for (i=0; i<bytes; i++)
			{
				MMC_REG(MMC_MMCTXFIFO) = *src++;
			}
			if (bytes < 32)
			{
				MMC_REG(MMC_MMCPRTBUF) = MMC_PRTBUF_BUF_PART_FULL;
			}
			len -= bytes;
		}
		status = MMC_REG(MMC_MMCSTAT);
		if (status & MMC_STAT_ERRORS)
		{
			printf("MMC_STAT error %lx\n", status);
			return -1;
		}
	}
	MMC_REG(MMC_MMCIMASK) = ~MMC_I_MASK_DATA_TRAN_DONE;
	while (!(MMC_REG(MMC_MMCIREG) & MMC_I_REG_DATA_TRAN_DONE));
	MMC_REG(MMC_MMCIMASK) = ~MMC_I_MASK_PRG_DONE;
	while (!(MMC_REG(MMC_MMCIREG) & MMC_I_REG_PRG_DONE));
	status = MMC_REG(MMC_MMCSTAT);
	if (status & MMC_STAT_ERRORS)
	{
		printf("MMC_STAT error %lx\n", status);
		return -1;
	}
	return 0;
}




int MmcRead(uchar *dst, ulong src, int size)
{
	ulong end, part_start, part_end, part_len, aligned_start, aligned_end;
	ulong mmc_block_size, mmc_block_address;

#ifdef MMC_DEBUG
	printf("MmcRead dst %lx src %lx len %d\n", (ulong)dst, src, size);
#endif

	if (size == 0)
	{
		return 0;
	}

	if (!mmc_ready)
	{
		printf("Please initial the MMC first\n");
		return -1;
	}

	mmc_block_size = MMC_BLOCK_SIZE;
	mmc_block_address = ~(mmc_block_size - 1);

	src -= CFG_MMC_BASE;
	end = src + size;
	part_start = ~mmc_block_address & src;
	part_end = ~mmc_block_address & end;
	aligned_start = mmc_block_address & src;
	aligned_end = mmc_block_address & end;

	/* all block aligned accesses */
#ifdef MMC_DEBUG
	printf("src %lx dst %lx end %lx pstart %lx pend %lx astart %lx aend %lx\n",
	src, (ulong)dst, end, part_start, part_end, aligned_start, aligned_end);
#endif
	if (part_start)
	{
		part_len = mmc_block_size - part_start;
#ifdef MMC_DEBUG
		printf("ps src %lx dst %lx end %lx pstart %lx pend %lx astart %lx aend %lx\n",
		src, (ulong)dst, end, part_start, part_end, aligned_start, aligned_end);
#endif
		if ((mmc_block_read(mmc_buf, aligned_start, mmc_block_size)) < 0)
		{
			return -1;
		}
		memcpy((char *)dst, (char *)(mmc_buf+part_start), part_len);
		dst += part_len;
		src += part_len;
	}
#ifdef MMC_DEBUG
	printf("src %lx dst %lx end %lx pstart %lx pend %lx astart %lx aend %lx\n",
	src, (ulong)dst, end, part_start, part_end, aligned_start, aligned_end);
#endif
	for (; src < aligned_end; src += mmc_block_size, dst += mmc_block_size)
	{
#ifdef MMC_DEBUG
		printf("al src %lx dst %lx end %lx pstart %lx pend %lx astart %lx aend %lx\n",
		src, (ulong)dst, end, part_start, part_end, aligned_start, aligned_end);
#endif
		if ((mmc_block_read((uchar *)(dst), src, mmc_block_size)) < 0)
		{
			return -1;
		}
	}
#ifdef MMC_DEBUG
	printf("src %lx dst %lx end %lx pstart %lx pend %lx astart %lx aend %lx\n",
	src, (ulong)dst, end, part_start, part_end, aligned_start, aligned_end);
#endif
	if (part_end && src < end)
	{
#ifdef MMC_DEBUG
		printf("pe src %lx dst %lx end %lx pstart %lx pend %lx astart %lx aend %lx\n",
		src, (ulong)dst, end, part_start, part_end, aligned_start, aligned_end);
#endif
		if ((mmc_block_read(mmc_buf, aligned_end, mmc_block_size)) < 0)
		{
			return -1;
		}
		memcpy((char *)dst, (char *)mmc_buf, part_end);
	}
	return size;
}

int MmcWrite(ulong dst, uchar *src, int size)
{
	ulong end, part_start, part_end, part_len, aligned_start, aligned_end;
	ulong mmc_block_size, mmc_block_address;

	if (size == 0)
	{
		return 0;
	}

	if (!mmc_ready)
	{
		printf("Please initial the MMC first\n");
		return -1;
	}

	mmc_block_size = MMC_BLOCK_SIZE;
	mmc_block_address = ~(mmc_block_size - 1);

	dst -= CFG_MMC_BASE;
	end = dst + size;
	part_start = ~mmc_block_address & dst;
	part_end = ~mmc_block_address & end;
	aligned_start = mmc_block_address & dst;
	aligned_end = mmc_block_address & end;

	/* all block aligned accesses */
#ifdef MMC_DEBUG
	printf("src %lx dst %lx end %lx pstart %lx pend %lx astart %lx aend %lx\n",
	src, (ulong)dst, end, part_start, part_end, aligned_start, aligned_end);
#endif
	if (part_start)
	{
		part_len = mmc_block_size - part_start;
#ifdef MMC_DEBUG
		printf("ps src %lx dst %lx end %lx pstart %lx pend %lx astart %lx aend %lx\n",
		(ulong)src, dst, end, part_start, part_end, aligned_start, aligned_end);
#endif
		if ((mmc_block_read(mmc_buf, aligned_start, mmc_block_size)) < 0)
		{
			return -1;
		}
		memcpy((char *)(mmc_buf+part_start), src, part_len);
		if ((mmc_block_write(aligned_start, mmc_buf, mmc_block_size)) < 0)
		{
			return -1;
		}
		dst += part_len;
		src += part_len;
	}
#ifdef MMC_DEBUG
	printf("src %lx dst %lx end %lx pstart %lx pend %lx astart %lx aend %lx\n",
	src, (ulong)dst, end, part_start, part_end, aligned_start, aligned_end);
#endif
	for (; dst < aligned_end; src += mmc_block_size, dst += mmc_block_size)
	{
#ifdef MMC_DEBUG
		printf("al src %lx dst %lx end %lx pstart %lx pend %lx astart %lx aend %lx\n",
		src, (ulong)dst, end, part_start, part_end, aligned_start, aligned_end);
#endif
		if ((mmc_block_write(dst, (uchar *)src, mmc_block_size)) < 0)
		{
			return -1;
		}
	}
#ifdef MMC_DEBUG
	printf("src %lx dst %lx end %lx pstart %lx pend %lx astart %lx aend %lx\n",
	src, (ulong)dst, end, part_start, part_end, aligned_start, aligned_end);
#endif
	if (part_end && dst < end)
	{
#ifdef MMC_DEBUG
		printf("pe src %lx dst %lx end %lx pstart %lx pend %lx astart %lx aend %lx\n",
		src, (ulong)dst, end, part_start, part_end, aligned_start, aligned_end);
#endif
		if ((mmc_block_read(mmc_buf, aligned_end, mmc_block_size)) < 0)
		{
			return -1;
		}
		memcpy((char *)mmc_buf, (char *)src, part_end);
		if ((mmc_block_write(aligned_end, mmc_buf, mmc_block_size)) < 0)
		{
			return -1;
		}
	}
	return size;
}

int MmcInit(int verbose)
{
 	int retries, rc = -ENODEV;
	uchar *resp;					// To store response.
	int Count;

	CLK_REG(CLK_CKEN) |= CKEN12_MMC; /* enable MMC unit clock */

	mmc_csd.c_size = 0;
	SDCardType = 0;
	mmc_ready = 0;
	RCA = MMC_DEFAULT_RCA;

	MMC_REG(MMC_MMCCLKRT)  = MMC_CLKRT_0_3125MHZ;
	MMC_REG(MMC_MMCRESTO)  = MMC_RES_TO_MAX;
	MMC_REG(MMC_MMCSPI)    = MMC_SPI_DISABLE;

	/* reset */
	retries = 10;
	resp = mmc_cmd(MMC_CMD_RESET, 0, 0, 0);

	// Assume the card is MMC Card.

	resp = mmc_cmd(MMC_CMD_SEND_OP_COND, 0x00ff, 0xc000, MMC_CMDAT_INIT|MMC_CMDAT_BUSY|MMC_CMDAT_R3);
	while (retries-- && resp && !(resp[4] & 0x80))
	{
#ifdef MMC_DEBUG
		printf("resp %x %x\n", resp[0], resp[1]);
#endif
		udelay(50);
		resp = mmc_cmd(MMC_CMD_SEND_OP_COND, 0x00ff, 0xff00, MMC_CMDAT_BUSY|MMC_CMDAT_R3);
	}

	if(!resp)
	{
#ifdef MMC_DEBUG
		printf("Not MMC, try SDCard commands\n");
#endif
		// Reset again.
		resp = mmc_cmd(MMC_CMD_RESET, 0, 0, 0);
		resp = mmc_cmd(MMC_CMD_SD_APPL, 0, 0, MMC_CMDAT_R1);
		udelay(10000);
		resp = mmc_cmd(MMC_CMD_SD_SEND_OP_COND, 0x0020, 0x0000, MMC_CMDAT_INIT|MMC_CMDAT_BUSY|MMC_CMDAT_R3);
		if(resp)
			SDCardType = 1;
		while (retries-- && resp && !(resp[4] & 0x80))
		{
#ifdef MMC_DEBUG
			printf("resp %x %x\n", resp[0], resp[1]);
#endif
			udelay(10000);
			resp = mmc_cmd(MMC_CMD_SD_APPL, 0, 0, MMC_CMDAT_R1);
			udelay(10000);
			resp = mmc_cmd(MMC_CMD_SD_SEND_OP_COND, 0x0020, 0x0000, MMC_CMDAT_BUSY|MMC_CMDAT_R3);
		}
	}

	/* try to get card id */
	resp = mmc_cmd(MMC_CMD_ALL_SEND_CID, 0, 0, MMC_CMDAT_R2);
	if (resp)
	{
		/* TODO configure mmc driver depending on card attributes */
		if(SDCardType)
		{
			memcpy((char *)&CID.sCID, resp, sizeof(mmc_sd_cid_t));
			printf("SD Card found.\n");
			if (verbose)
			{
				printf("Card desciption is:\n");
				printf("Manufacturer ID = %02x\n", CID.sCID.oid);
				printf("HW/FW Revision = %x %x\n",CID.sCID.hwrev, CID.sCID.fwrev);
				printf("Product Name = ");
				for(Count = 5; Count >=0; Count--)
					printf("%c", CID.sCID.name[Count]);
				printf("\n");
	
				printf("Serial Number = %02x%02x%02x%02x\n",
								CID.sCID.psn[3], CID.sCID.psn[2], CID.sCID.psn[1], CID.sCID.psn[0]);
				printf("Month = %d\n",CID.sCID.month);
				printf("Year = %d\n",1997 + CID.sCID.year);
			}
		}
		else
		{
			memcpy((char *)&CID.mCID, resp, sizeof(mmc_sd_cid_t));
			printf("MMC Card found.\n");
			if (verbose)
			{
				printf("Card desciption is:\n");
				printf("Manufacturer ID = %02x%02x\n", CID.mCID.oid[0], CID.mCID.oid[1]);
				printf("HW/FW Revision = %x %x\n",CID.mCID.hwrev, CID.mCID.fwrev);
				printf("Product Name = ");
				for(Count = 5; Count >=0; Count--)
					printf("%c", CID.mCID.name[Count]);
				printf("\n");
	
				printf("Serial Number = %02x%02x%02x%02x\n",
								CID.mCID.psn[3], CID.mCID.psn[2], CID.mCID.psn[1], CID.mCID.psn[0]);
				printf("Month = %d\n",CID.mCID.month);
				printf("Year = %d\n",1997 + CID.mCID.year);
			}
		}


		/* SD or MMC exists, get CSD too */
		if(SDCardType)
		{
			resp = mmc_cmd(MMC_CMD_SET_RCA, 0, 0, MMC_CMDAT_R1);
			RCA = resp[4] << 8 | resp[3];
		}
		else
		{
			resp = mmc_cmd(MMC_CMD_SET_RCA, RCA, 0, MMC_CMDAT_R1);
		}

		if(SDCardType)
		{
			resp = mmc_cmd(MMC_CMD_SEND_CSD, RCA, 0, MMC_CMDAT_R2);
		}
		else
		{
			resp = mmc_cmd(MMC_CMD_SEND_CSD, RCA, 0, MMC_CMDAT_R2);
		}

		if (resp)
		{
			mmc_csd_t *csd = (mmc_csd_t *)resp;
			memcpy((char *)&mmc_csd, (char *)csd, sizeof(mmc_csd_t));
			rc = 0;
			mmc_ready = 1;
			if (verbose)
			{
				printf("c_size = %d\n",mmc_csd.c_size);
				printf("c_size_mult = %d\n",mmc_csd.c_size_mult1);
			}

			// Set transfer speed to 20 MHz.
			MMC_REG(MMC_MMCCLKRT) = 0;	/* 20 MHz */
			resp = mmc_cmd(MMC_CMD_SEL_CARD, RCA, 0, MMC_CMDAT_R1);

		}
		else
		{	mmc_ready = 0;
		}
	}

#if INCLUDE_FS
	if(mmc_ready)
	{
		// Fill in mmc_fsops structure.
		if(SDCardType)
			strcpy(mmc_fsops.name, "SDCard");
		else
			strcpy(mmc_fsops.name, "MMC");
		mmc_fsops.start_address = CFG_MMC_BASE;
		mmc_fsops.media_read = MmcRead;
		mmc_fsops.media_write = MmcWrite;

		rc = mount_fs(&mmc_fsops);
	}
#endif

	return rc;
}


char *mmcHelp[] = {
	"mmc operations",
	"{op} [args]",
	"Ops...",
	"  init",
	"  write {dest} {src} {byte_cnt}",
	"  read {dest} {src} {byte_cnt}",
	"",
	"",
	0,
};

/* MmcCmd():
 *	Code that handles the user interface.  See mmcHelp[] above for usage.
 */
int
MmcCmd(int argc,char *argv[])
{
	int		snum, ret;
	ulong	dest, src, oints;
	long	bytecnt, rslt;

	ret = CMD_SUCCESS;

	if (strcmp(argv[1],"init") == 0)
		ret = MmcInit(1);
	else if (!strcmp(argv[1],"write")) 
	{
		if (argc == 5)
		{
			dest = strtoul(argv[2],(char **)0,0);
			src = strtoul(argv[3],(char **)0,0);
			bytecnt = (long)strtoul(argv[4],(char **)0,0);
			rslt = MmcWrite((ulong)dest,(uchar *)src,bytecnt);
			if (rslt == -1)
			{
				printf("Write failed\n");
				ret = CMD_FAILURE;
			}
		}
		else
		{
			ret = CMD_PARAM_ERROR;
		}
	}
	else if (!strcmp(argv[1],"read")) 
	{
		if (argc == 5)
		{
			dest = strtoul(argv[2],(char **)0,0);
			src = strtoul(argv[3],(char **)0,0);
			bytecnt = (long)strtoul(argv[4],(char **)0,0);
			rslt = MmcRead((uchar *)dest,(ulong )src,bytecnt);
			if (rslt == -1)
			{
				printf("Read failed\n");
				ret = CMD_FAILURE;
			}
		}
		else
		{
			ret = CMD_PARAM_ERROR;
		}
	}
	else
	{
		ret = CMD_PARAM_ERROR;
	}

	return(ret);
}


#endif	// INCLUDE_MMC
