//==========================================================================
//
//      pxa250_lcd.c
//
// Author(s):   Michael Kelly - Cogent Computer Systems, Inc.
// Date:        05-27-2002
// Description:	Init Code for PXA250 LCD Controller
//
//
// Modified by Technology Rescue.
//	11.03.2003 : changed color depth.
//	12.03.2003 : changed frame buffer location for Windows CE.
//	15.04.2003 : added lcd_clr_scr func.
//	07.05.2003 : changed descriptor addresses and frame buffer address.
//	09.05.2003 : moved defines to board include file.
//	03.06.2003 : added lcd_switch_buffer.
//	11.06.2003 : removed dis/enable LCD in lcd_switch_buffer func.
//	01.10.2003 : fixed PAL bit in LCD_PDR(0xc) according to lcd_color_depth.
//	07.01.2004 : added new frame buffer address at the end of lcd_switch_buffer.
//
//==========================================================================

#include "config.h"
#include "cpuio.h"
#include "stddefs.h"
#include "genlib.h"
#if FONT_HEIGHT == 20
#include "font8x20.h"
#else
#include "font8x16.h"
#endif
#include "lcd_lut.h"

//--------------------------------------------------------------------------
// function prototypes and externs
//
int lcd_init(void);
void lcd_on(void);
void lcd_off(void);

void lcd_putchar(char c);
void lcd_writechar(uchar c);
void lcd_clr_row(int char_row);
void lcd_clr_scr(void);
void lcd_switch_buffer(void);

extern void udelay(int);
extern int GPIO_set(int);
extern int GPIO_clr(int);
extern int GPIO_tst(int);

// globals to keep track of foreground, background colors and x,y position
int lcd_color_depth;		// 4, 8 or 16
int lcd_fg_color;			// 0 to 15, used as lookup into VGA color table
int lcd_bg_color;			// 0 to 15, used as lookup into VGA color table
int lcd_col;				// current column, 0 to COLS_PER_SCREEN - 1
int lcd_row;				// current row, 0 to (ROWS_PER_SCREEN * 2) - 1
int lcd_tst_mode = 0;
ulong lcd_fb_offset;	   	// current offset into frame buffer for lcd_putchar

#define LCD_FB_SIZE(_depth_)				(((PIXELS_PER_COL * PIXELS_PER_ROW) * _depth_) / 8)

#define LCD_ROW_SIZE(_depth_)				((PIXELS_PER_ROW * _depth_) / 8)

#define LCD_GET_ADD(_row_, _col_, _depth_)	(((((_row_ * PIXELS_PER_ROW) * FONT_HEIGHT) \
											  + (_col_ * FONT_WIDTH)) \
											  * _depth_) / 8)

const ushort vga_lookup[] = {
LU_BLACK,		
LU_RED,			
LU_GREEN,		
LU_YELLOW,		
LU_BLUE,			
LU_VIOLET,		
LU_CYAN,			
LU_GREY,			
LU_WHITE,		
LU_BRT_RED,		
LU_BRT_GREEN,	
LU_BRT_YELLOW,
LU_BRT_BLUE,		
LU_BRT_VIOLET,
LU_BRT_CYAN,		
LU_BRT_WHITE	
};

//--------------------------------------------------------------------------
// lcd_init
//
// This function sets up the PXA250 LCD controller
//
int lcd_init()
{
	ulong index;
	int i;
	ushort temp16;

#ifdef LCD_DBG
	lcd_tst_mode = 1;
	printf("PXA250 LCD Initialization.\n");
	lcd_tst_mode = 0;
#endif

	lcd_color_depth = BIT_PER_PIXEL;

	index = 0;
	// Fill the palette buffer with the lookup table data
	for (i = 0; i < 256; i ++)
	{
		LCD_PAL(index) = (((lcd_lut[i][0] & 0xf8) << 8)		// Red
					    | ((lcd_lut[i][1] & 0xfc) << 3) 	// Green
					    | ((lcd_lut[i][2]) >> 3));			// Blue

		index += 2;
	}

	// update palette buffer with vga 16 color palette.
	for (i = 0, index = 0; i < 16; i ++, index+=2)
	{
		LCD_PAL(index) = vga_lookup[i];
	}

	// initialize the globals	
	lcd_fg_color = LCD_FG_DEF;
	lcd_bg_color = LCD_BG_DEF;
	lcd_row = 0;
	lcd_col = 0;
	lcd_fb_offset = 0x00;

	// fill the screen with the background color, remember for the console driver
	// we do double buffering, so fill twice as mush as the active screen size
	switch (lcd_color_depth){
		case 4:  temp16 = lcd_bg_color | lcd_bg_color << 4 | lcd_bg_color << 8 | lcd_bg_color << 12; break;
		case 8:  temp16 = lcd_bg_color | lcd_bg_color << 8; break;
		default: temp16 = lcd_bg_color; break;
	}
	for (i = 0; i < (PIXELS_PER_COL * PIXELS_PER_ROW) * 2; i += 2){	
		LCD_BUF(i) = temp16;
	}

	// kick off the frame descriptors
	// Setup the Palette Descriptor
	LCD_PDR(0x0) = LCD_FDR_ADD;								// point to the frame buffer descriptor
	LCD_PDR(0x4) = LCD_PAL_ADD;								// beginning of Palette Buffer in Memory
	LCD_PDR(0x8) = 0;										// Frame ID

	switch (lcd_color_depth)
	{
		case 4:
			LCD_PDR(0xc) = BIT26 | 32;						// PAL = 1, Frame Length in bytes = 32.
			break;
		case 8:
			LCD_PDR(0xc) = BIT26 | 512;						// PAL = 1, Frame Length in bytes = 512.
			break;
		default:
			LCD_PDR(0xc) = LCD_FB_SIZE(lcd_color_depth);	// 16-bit LCD doesn't need palette buffer.
			break;
	} // switch color depth

	// Setup the Frame Descriptor
	LCD_FDR(0x0) = LCD_FDR_ADD;								// point it to itself
	LCD_FDR(0x4) = LCD_BUF_ADD;								// beginning of Frame Buffer in Memory
	LCD_FDR(0x8) = 1;										// Frame ID
	LCD_FDR(0xc) = LCD_FB_SIZE(lcd_color_depth);			// Frame Length in bytes

	return 0;
}

//--------------------------------------------------------------------------
// lcd_on
//
// This function turns on the PXA250 LCD Controller and the Backlight Inverter
//
void lcd_on()
{
	// Enable the Backlight
	GPIO_set(LCD_BKL);

	// Setup Timing Registers
	// Note, LCD pixel clock is derived from MEMCLK which
	// defaults to 99.5Mhz.  Any other value requires changing
	// LCCR3_PCD.
	LCD_REG(LCD_LCCR3) = LCCR3_VAL;
	LCD_REG(LCD_LCCR2) = LCCR2_VAL;
	LCD_REG(LCD_LCCR1) = LCCR1_VAL;

	// Setup the final LCD control register, but
	// leave the controller off.  It will be controlled 
	// by lcd_on() and lcd_off() which also handles the
	// backlight
	LCD_REG(LCD_LCCR0) = LCCR0_VAL;


	// For color depths of 4 or 8 we point the PXA250 to 
	// the palette buffer descriptor, which will then point
	// to the frame buffer descriptor, while for 16 bpp 
	// we point directly to the frame buffer descriptor,
	// thus skipping the palette buffer descriptor.
	if ((lcd_color_depth == 4) || (lcd_color_depth == 8)){
		LCD_REG(LCD_FDADR0) = LCD_PDR_ADD;
	}
	else {
		LCD_REG(LCD_FDADR0) = LCD_FDR_ADD;
	}

	// Turn on the LCD Controller
	LCD_REG(LCD_LCCR0) = (LCD_REG(LCD_LCCR0) | LCCR0_ENB); 

	// wait for power-up sequence to complete
	udelay(10000);
}

//--------------------------------------------------------------------------
// lcd_off
//
// This function turns off the LCD Controller and the LCD backlight
//
void lcd_off()
{
	// Turn off the LCD Controller using the diable bit
	LCD_REG(LCD_LCCR0) |= (LCD_REG(LCD_LCCR0) | LCCR0_DIS);
	
	// Wait for the Enable Bit to go clear
	while(LCD_REG(LCD_LCCR0) & LCCR0_ENB) {}

	// Turn off the backlight
	GPIO_clr(LCD_BKL);
}

char *lcd_tstHelp[] = {
	" This command allows the user to test ",
	" the PXA250 LCD Controller.  The user ",
	" may set the color depth to run the ",
	" test at.  Note that the frame buffer ",
	" R/W test will test all of the frame ",
	" buffer regardless of depth.\n",
    " Usage:",
    " lcd_tst -[p,n,x,d:[4,8,16]]",
    " Options...",
	" -p run palette test",
    " -n run test without keycheck - CAUTION: RESET SYSTEM TO STOP!",
	" -d4 run test, force a depth of 4-bits/pixel",
	" -d8 run test, force a depth of 8-bits/pixel",
	" -d16 run test, force a depth of 16-bits/pixel",
	" -x init only, do not run frame buffer tests",
	" no options, default to current mode and depth",
	0
};

int lcd_tst(int argc,char *argv[])
{
	volatile ushort wr16, rd16;
	int i, x, color, maxcolor, opt;
	int no_wait = 0;
	int init_only = 0;
	int palette_only = 0;
	char c;

	lcd_tst_mode = 1;
	lcd_off();

    while ((opt=getopt(argc,argv,"pclnsxd:4,8,16")) != -1) {
        switch(opt) {
			case 'd':	// set the color depth
                switch(*optarg) {
            	    case '4':
	        	        lcd_color_depth = 4;
						printf("Forcing 4bpp Mode!\n");
        	            break;
    	            case '8':
            	        lcd_color_depth = 8;
						printf("Forcing 8bpp Mode!\n");
                    	break;
	                default:	// test with 16bpp
	                    lcd_color_depth = 16;
						printf("Forcing 16bpp Mode!\n");
						break;
                }
	            break;
	        case 'n':   // no waiting for keypress - fastest operation
				no_wait = 1;
				printf("No Keypress Mode, Must Reset System to Stop!\n");
				break;
	        case 'x':   // init only 
				no_wait = 1;
				printf("Initializing LCD, Skipping testsp!\n");
				init_only = 1;
				break;
			case 'p':	// palette only.
				palette_only = 1;
			default:	// test with current mode
				break;
		}
	}

	// get the new parameters into the LCD controller
	lcd_init();
	lcd_on();

	if(!palette_only)
	{
		if (init_only) return 0;
	
		printf("Frame Buffer R/W...");
		// do an address=data read/write test on the frame buffer
		// PIXELS_PER_COL * PIXELS_PER_ROW is the highest pixel.  
		// Multiply by bits_per_pixel (sed_color_depth), then
		// divide by 8 to get the actual byte count.
		for (i = 0; i < LCD_FB_SIZE(lcd_color_depth); i += 2){	
			LCD_BUF(i) = i & 0xffff;
			rd16 = LCD_BUF(i);
			if(rd16 != (i & 0xffff)){
				printf("Fail at 0x%08x, WR 0x%08x, RD 0x%04x!\n",LCD_BUF_ADD + i, i, rd16);
				return -1;
			}
		} 
	
		printf("OK!, Press key to continue.\n");
	
		c = getchar();
	}

	printf("Palette test...");

	wr16 = 0;
	color = 0;
	switch (lcd_color_depth)
	{
		case 4:  maxcolor = 15;
				 break;
		case 8:  maxcolor = 255;
				 break;
		default: maxcolor = 65535;
				 break;
	}
	
	for (i = 0; i < LCD_FB_SIZE(lcd_color_depth); i += 20)
	{
		switch (lcd_color_depth)
		{
			case 4:  wr16 = color | color << 4 | color << 8 | color << 12; break;
			case 8:  wr16 = color | color << 8; break;
			default: wr16 = vga_lookup[color]; break;	// 16-bits bypasses the lookup table
		}
		for(x=0; x<40; x+=2)
		{
			LCD_BUF(i+x) = wr16;
		}
		if(color>=maxcolor)
			color = 0;
		else
			color++;
	}

	c = getchar();

	if(!palette_only)
	{
		printf("Frame Buffer Start: 0x%08x, End 0x%08x\n",LCD_BUF_ADD, LCD_BUF_ADD + LCD_FB_SIZE(lcd_color_depth));
		if (no_wait)
		{
			printf("Begin Full Screen Color Test.\n");
			while(1){
				// fill the frame buffer with incrementing color values
				for (x = 0; x < 16; x++){
					switch (lcd_color_depth){
						case 4:  wr16 = x | x << 4 | x << 8 | x << 12; break;
						case 8:  wr16 = x | x << 8; break;
						default: wr16 = vga_lookup[x]; break;	// 16-bits bypasses the lookup table
					}
					for (i = 0; i < LCD_FB_SIZE(lcd_color_depth); i += 2){	
						LCD_BUF(i) = wr16;
					}
				} // for x
			} // while
		} // no_wait
		else
		{
			printf("Begin Full Screen Color Test, Press any key to go to next color, \'x\' to end.\n");
			while(1){
				// fill the frame buffer with incrementing color values
				for (x = 0; x < 16; x++){
					switch (lcd_color_depth){
						case 4:  wr16 = x | x << 4 | x << 8 | x << 12; break;
						case 8:  wr16 = x | x << 8; break;
						default: wr16 = vga_lookup[x]; break;	// 16-bits bypasses the lookup table
					}
					for (i = 0; i < LCD_FB_SIZE(lcd_color_depth); i += 2){	
						LCD_BUF(i) = wr16;
					}
					c = getchar();
					if (c == 'x') goto lcd_tst_next;
				} // for x
			} // while
		} // else no keycheck test
	

	}

lcd_tst_next:
	lcd_off();
	lcd_init();
	lcd_on();

	lcd_tst_mode = 0;
	return 0;
}

// lcd_scroll()
//
// Because we are most likely running out of FLASH and probably also with
// cache disabled, a brute force memcpy of the whole screen would be very
// slow, even with reduced color depths.  Instead, we define a frame buffer
// that is twice the size of our actual display.
//
// With the double buffering, we always write to the first buffer, even
// when the second buffer is active.  This allows us to scroll by adjusting
// the starting and ending addresses in the PXA250 by one row.  When we
// reach the end of our virtual buffer, we reset the starting and ending
// addresses to the first buffer.
//
void lcd_scroll(void) {

	lcd_row++;

	// clear the new row(s)
	lcd_clr_row(lcd_row);
	if (lcd_row > (ROWS_PER_SCREEN - 1)){
		lcd_clr_row(lcd_row - ROWS_PER_SCREEN);
	}
	// when lcd_y_pos is greater than ROWS_PER_SCREEN we just adjust the
	// start and end addresses in the PXA250.  If it is equal to 2 *
	// ROWS_PER_SCREEN, we reset the start and end addresses to LCD_MEM_BASE.
	if (lcd_row > (ROWS_PER_SCREEN - 1))
	{
		if (lcd_row > ((ROWS_PER_SCREEN * 2) - 1))
		{
			lcd_fb_offset = 0x00;
			lcd_row = ROWS_PER_SCREEN - 1;
		}
		else
		{
			// calculate the new offset address of the frame buffer in words
			lcd_fb_offset += (LCD_GET_ADD(1, 0, lcd_color_depth));
		}

#ifdef LCD_DBG
	lcd_tst_mode = 1;
	printf("LCD New FB Offset = 0x%08lx.\n ", lcd_fb_offset);
	lcd_tst_mode = 0;
#endif
		LCD_FDR(0x4) = LCD_BUF_ADD + lcd_fb_offset;				// new beginning of Frame Buffer in Memory

	} // if (lcd_row > (ROWS_PER_SCREEN - 1))
	return;
		

} // 

//--------------------------------------------------------------------------
// lcd_clr_row()
//
// This routine writes the background color to the font row specified
//
void lcd_clr_row(int char_row){

	ulong lcd_mem_add;
	int i;
	ushort wr16;

	// clear the desired row
	lcd_mem_add = LCD_GET_ADD(char_row, 0, lcd_color_depth);

#ifdef LCD_DBG
	lcd_tst_mode = 1;
	printf("LCD Clear Row %d, FB Add 0x%08lx.\n ", char_row, lcd_mem_add);
	lcd_tst_mode = 0;
#endif

	switch (lcd_color_depth)
	{
		case 4:
			wr16 = ((lcd_bg_color << 12) | (lcd_bg_color << 8) | (lcd_bg_color << 4) | (lcd_bg_color << 0));
			for (i = 0; i < ((PIXELS_PER_ROW * FONT_HEIGHT) / 2); i += 2){
				LCD_BUF(lcd_mem_add) = wr16;
				lcd_mem_add += 2;
			} // for font_row
			break;
		case 8:
			wr16 = ((lcd_bg_color << 8) | (lcd_bg_color << 0));
			for (i = 0; i < (PIXELS_PER_ROW * FONT_HEIGHT); i += 2){
				LCD_BUF(lcd_mem_add) = wr16;
				lcd_mem_add += 2;
			} // for font_row
			break;
		case 16:
			wr16 = (lcd_bg_color << 0);
			for (i = 0; i < ((PIXELS_PER_ROW * FONT_HEIGHT) * 2); i += 2){
				LCD_BUF(lcd_mem_add) = wr16;
				lcd_mem_add += 2;
			} // for font_row
			break;
	} // switch lcd_color_depth
} // lcd_clr_row


//--------------------------------------------------------------------------
// lcd_putchar()
//
// This routine parses the character and calls lcd_writechar if it is a
// printable character
//
void lcd_putchar(char c){

	if (lcd_tst_mode) return;

	// First parse the character to see if it is printable or an acceptable control
    switch (c) {
	    case '\r':
	        lcd_col = 0;
	        return;
	    case '\n':
	        lcd_scroll();
	        return;
	    case '\b':
	        lcd_col--;
	        if (lcd_col < 0)
	        {
	            lcd_row--;
	            if (lcd_row < 0) lcd_row = 0;
	            lcd_col = COLS_PER_SCREEN - 1;
		    }
			c = 0;		// erase the character
			lcd_writechar(c);
	        break;
	    default:
	        if (((uchar)c < FIRST_CHAR) || ((uchar)c > LAST_CHAR)) return; // drop anything we can't print
			c -= FIRST_CHAR;	// get aligned to the first printable character
			lcd_writechar(c);
			// advance to next column
			lcd_col++;
		    if (lcd_col == COLS_PER_SCREEN)
		    {
				lcd_col = 0;
			    lcd_scroll();
			}
			break;
    } 

} // lcd_putchar()

//--------------------------------------------------------------------------
// lcd_writechar()
//
// This routine writes the character to the screen at the current cursor
// location.
//
void lcd_writechar(uchar c) {

	ulong lcd_mem_add;
	int font_row, font_col;
	uchar font_data8;
	ushort wr16;

	// Convert the current row,col and color depth values 
	// into an address
	lcd_mem_add = LCD_GET_ADD(lcd_row, lcd_col, lcd_color_depth);

#ifdef LCD_DBG
	lcd_tst_mode = 1;
	printf("SED writechar at row %d, col %d, FB Add 0x%08lx.\n ", lcd_row, lcd_col, lcd_mem_add);
	lcd_tst_mode = 0;
#endif

	if (FONT_WIDTH == 8)
	{
		switch (lcd_color_depth){
			case 4:
				// Now render the font by painting one font row at a time
				for (font_row = 0; font_row < FONT_HEIGHT; font_row++){
					// get the font row of data
					font_data8 = fonttbl[(c * FONT_HEIGHT) + font_row];
					for (font_col = 0; font_col < 8; font_col += 4)
					{
						// get a words worth of pixels
						wr16 = (((font_data8 & 0x80) ? lcd_fg_color << 0 : lcd_bg_color << 0)
							  | ((font_data8 & 0x40) ? lcd_fg_color << 4 : lcd_bg_color << 4)
							  | ((font_data8 & 0x20) ? lcd_fg_color << 8 : lcd_bg_color << 8)
							  | ((font_data8 & 0x10) ? lcd_fg_color << 12 : lcd_bg_color << 12));
						font_data8 = font_data8 << 4;
//						wr16 = (((font_data8 & 0x80) ? lcd_fg_color << 12 : lcd_bg_color << 12)
//							  | ((font_data8 & 0x40) ? lcd_fg_color << 8 : lcd_bg_color << 8)
//							  | ((font_data8 & 0x20) ? lcd_fg_color << 4 : lcd_bg_color << 4)
//							  | ((font_data8 & 0x10) ? lcd_fg_color << 0 : lcd_bg_color << 0));
//						font_data8 = font_data8 << 4;
						LCD_BUF(lcd_mem_add) = wr16;
						// if we are in the 2nd frame buffer, write to the 1st frame buffer also
						if (lcd_row > (ROWS_PER_SCREEN - 1)){
							LCD_BUF(lcd_mem_add - LCD_FB_SIZE(lcd_color_depth)) = wr16;
						}
						lcd_mem_add += 2;
					} // for font_col
					// go to the next pixel row
					lcd_mem_add += (LCD_ROW_SIZE(lcd_color_depth) - ((FONT_WIDTH * lcd_color_depth) / 8));
				} // for font_row
				break;

			case 8:
				// Now render the font by painting one font row at a time
				for (font_row = 0; font_row < FONT_HEIGHT; font_row++){
					// get the font row of data
					font_data8 = fonttbl[(c * FONT_HEIGHT) + font_row];
					for (font_col = 0; font_col < 8; font_col += 2)
					{
						// get a words worth of pixels
						wr16 = (((font_data8 & 0x80) ? lcd_fg_color << 0 : lcd_bg_color << 0)
							  | ((font_data8 & 0x40) ? lcd_fg_color << 8 : lcd_bg_color << 8));
						font_data8 = font_data8 << 2;
//						wr16 = (((font_data8 & 0x80) ? lcd_fg_color << 8 : lcd_bg_color << 8)
//							  | ((font_data8 & 0x40) ? lcd_fg_color << 0 : lcd_bg_color << 0));
//						font_data8 = font_data8 << 2;
						LCD_BUF(lcd_mem_add) = wr16;
						// if we are in the 2nd frame buffer, write to the 1st frame buffer also
						if (lcd_row > (ROWS_PER_SCREEN - 1)){
							LCD_BUF(lcd_mem_add - LCD_FB_SIZE(lcd_color_depth)) = wr16;
						}
						lcd_mem_add += 2;
					} // for font_col
					// go to the next pixel row
					lcd_mem_add += (LCD_ROW_SIZE(lcd_color_depth) - ((FONT_WIDTH * lcd_color_depth) / 8));
				} // for font_row
				break;

			default:	// 16bpp is the default
				// Now render the font by painting one font row at a time
				for (font_row = 0; font_row < FONT_HEIGHT; font_row++){
					// get the font row of data
					font_data8 = fonttbl[(c * FONT_HEIGHT) + font_row];
					for (font_col = 0; font_col < 8; font_col++)
					{
						// get a words worth of pixels
						wr16 = ((font_data8 & 0x80) ? lcd_fg_color : lcd_bg_color);
						font_data8 = font_data8 << 1;
						LCD_BUF(lcd_mem_add) = wr16;
						// if we are in the 2nd frame buffer, write to the 1st frame buffer also
						if (lcd_row > (ROWS_PER_SCREEN - 1)){
							LCD_BUF(lcd_mem_add - LCD_FB_SIZE(lcd_color_depth)) = wr16;
						}
						lcd_mem_add += 2;
					} // for font_col
					// go to the next pixel row
					lcd_mem_add += (LCD_ROW_SIZE(lcd_color_depth) - ((FONT_WIDTH * lcd_color_depth) / 8));
				} // for font_row
				break;
		} // switch lcd_color depth
	} // FONT_WIDTH == 8
	else
	{
		return;
	}
} // lcd_writechar()	


void lcd_clr_scr(void)
{
	int	row;

	/* Clear screen by clearing each lines. */
	for(row = 0; row < (ROWS_PER_SCREEN * 2); row++)
	{
		lcd_clr_row(row);
	}

	/* Reset cursor position. */
	lcd_row = 0;
	lcd_col = 0;
	lcd_fb_offset = 0x00;

}

/* clear_memory():
 * Clear n dwords memory from start address sp.
 * Return sp.
 */
int *
clear_memory(register int *sp, register int n)
{
	register int *sp0 = sp;

	while (--n >= 0)
		*sp++ = 0x0;
	return (sp0);
}

/*
	lcd_switch_buffer:
		This function move 1st part of screen buffer to second part and
		changes lcd frame buffer start address to 2nd part.
		Don't disable lcd controller to prevent from pixel shift.
*/
void lcd_switch_buffer(void)
{
	ulong index;
	int i;

	/* Move LCD frame buffer to windows CE frame buffer. */
	memcpy((char *)LCD_BUF_ADD_CE, (char *)(LCD_FDR(0x04)), LCD_FB_SIZE(lcd_color_depth));

	/* Update palette buffer with vga 16 color palette. */
	for (i = 0, index = 0; i < 16; i ++, index+=2)
	{
		LCD_PAL_CE(index) = vga_lookup[i];
	}

	/* Move DMA descriptors to Windows CE address. */
	// Setup the Palette Descriptor
	LCD_PDR_CE(0x0) = LCD_FDR_ADD_CE;						// point to the frame buffer descriptor
	LCD_PDR_CE(0x4) = LCD_PAL_ADD_CE;						// beginning of Palette Buffer in Memory
	LCD_PDR_CE(0x8) = 0;									// Frame ID

	switch (lcd_color_depth)
	{
		case 4:
			LCD_PDR_CE(0xc) = BIT26 | 32;					// PAL = 1, Frame Length in bytes = 32.
			break;
		case 8:
			LCD_PDR_CE(0xc) = BIT26 | 512;					// PAL = 1, Frame Length in bytes = 512.
			break;
		default:
			LCD_PDR_CE(0xc) = LCD_FB_SIZE(lcd_color_depth);	// 16-bit LCD doesn't need palette buffer.
			break;
	} // switch color depth

	// Setup the Frame Descriptor
	LCD_FDR_CE(0x0) = LCD_FDR_ADD_CE;						// point it to itself
	LCD_FDR_CE(0x4) = LCD_BUF_ADD_CE;						// beginning of Frame Buffer in Memory
	LCD_FDR_CE(0x8) = 1;									// Frame ID
	LCD_FDR_CE(0xc) = LCD_FB_SIZE(lcd_color_depth);			// Frame Length in bytes

#ifdef RE_INIT_LCD

	/* Disable LCD controller by setting DIS bit. */
	LCD_REG(LCD_LCCR0) |= LCCR0_DIS;

	/* Wait for the Enable Bit to go clear. */
	while(LCD_REG(LCD_LCCR0) & LCCR0_ENB) {}

	/* Re-enable LCD controller by resetting DIS bit.
	   This should be done immediately to prevent from seeing LCD switching.
	 */
	LCD_REG(LCD_LCCR0) &= ~LCCR0_DIS;
	LCD_REG(LCD_LCCR0) |= LCCR0_ENB;

	/* Update LCD_FDADR0 reg. */
	if ((lcd_color_depth == 4) || (lcd_color_depth == 8)){
		LCD_REG(LCD_FDADR0) = LCD_PDR_ADD_CE;
	}
	else {
		LCD_REG(LCD_FDADR0) = LCD_FDR_ADD_CE;
	}

#else

	if ((lcd_color_depth == 4) || (lcd_color_depth == 8))
	{
		LCD_PDR(0x0) = LCD_PDR_ADD_CE;						// next palette desc.
		LCD_PDR(0x4) = LCD_BUF_ADD_CE;						// new beginning of Frame Buffer in Memory
	}

	LCD_FDR(0x0) = LCD_FDR_ADD_CE;							// next frame desc.
	LCD_FDR(0x4) = LCD_BUF_ADD_CE;							// new beginning of Frame Buffer in Memory

	lcd_tst_mode = 1;										// Prevent umon from writing to old lcd buffer.

	/* Clear area between old frame buffer and new frame descriptor. */
	clear_memory((int *)LCD_BUF_ADD, (LCD_FDR_ADD_CE - LCD_BUF_ADD)/sizeof(int));


#endif

}



