//=============================================================================
//
//      etherdev.c
//
//      Ethernet Abstraction Layer for Micro Monitor
//
// Author(s):    Michael Kelly, Cogent Computer Systems, Inc.
// Date:         07/23/2003
// Description:  This file contains the interface layer between Micro Monitor
//               and the Ethernet driver for the AT91SAM9263 on the CSB737.
//
//=============================================================================

#include "config.h"
#include "cpuio.h"
#include "genlib.h"
#include "stddefs.h"
#include "ether.h"
#include "at91sam9263.h"
#include "at91sam9263_emac.h"

#if INCLUDE_ETHERNET

extern void udelay(int);

// Receive Buffer Descriptor Header
typedef struct
{
	ulong address;		// receive buffer address
	ulong status;		// status
} RXBUF_HDR;

// Bit assignments for Receive Buffer Descriptor *CSB637*
// Address - Word 0
//#define RXBUF_ADD_BASE_MASK			0xfffffffc	// Base address of the receive buffer in memory
//#define RXBUF_ADD_WRAP				BIT1		// If this bit is set, this is the last buffer in the ring
//#define RXBUF_ADD_OWNED				BIT0		// 1 = SW owns the pointer, 0 = MAC owns buffer

// Status - Word 1
//#define RXBUF_STAT_BCAST			BIT31		// Global all ones broadcast address detected
//#define RXBUF_STAT_MULTI			BIT30		// Multicast hash match
//#define RXBUF_STAT_UNI				BIT29		// Unicast hash match
//#define RXBUF_STAT_EXT				BIT28		// External address (optional)
//#define RXBUF_STAT_UNK				BIT27		// Unknown source address (reserved for future use)
//#define RXBUF_STAT_LOC1				BIT26		// Local address match (Specific address 1 match)
//#define RXBUF_STAT_LOC2				BIT25		// Local address match (Specific address 2 match)
//#define RXBUF_STAT_LOC3				BIT24		// Local address match (Specific address 3 match)
//#define RXBUF_STAT_LOC4				BIT23		// Local address match (Specific address 4 match)
//#define RXBUF_STAT_LEN_MASK			0x3ff		// Length of frame including FCS

// Bit assignments for Receive Buffer Descriptor *CSB737*
// Address - Word 0
#define RXBUF_ADD_BASE_MASK			0xfffffffc		// Base address of the receive buffer in memory
#define RXBUF_ADD_WRAP				BIT1			// If this bit is set, this is the last buffer in the ring
#define RXBUF_ADD_OWNED				BIT0			// 1 = SW owns the pointer, 0 = MAC owns buffer

// Status - Word 1
#define RXBUF_STAT_BCAST			BIT31			// Global all ones broadcast address detected
#define RXBUF_STAT_MULTI			BIT30			// Multicast hash match
#define RXBUF_STAT_UNI				BIT29			// Unicast hash match
#define RXBUF_STAT_EXT				BIT28			// External address (optional)
#define RXBUF_STAT_UNK				BIT27			// Unknown source address (reserved for future use)
#define RXBUF_STAT_LOC1				BIT26			// Local address match (Specific address 1 match)
#define RXBUF_STAT_LOC2				BIT25			// Local address match (Specific address 2 match)
#define RXBUF_STAT_LOC3				BIT24			// Local address match (Specific address 3 match)
#define RXBUF_STAT_LOC4				BIT23			// Local address match (Specific address 4 match)
#define RXBUF_STAT_MATCH			BIT22			// Type ID Match
#define RXBUF_STAT_VLANT			BIT21			// VLAN tag detected (i.e., type id of 0x8100)
#define RXBUF_STAT_PRIOT			BIT20			// Priority tag detected (i.e., type id of 0x8100 and null VLAN identifier)
#define RXBUF_STAT_VLANP			((_x_ & 0x7) << 17)	// VLAN Priority (only valid if bit 21 is set)
#define RXBUF_STAT_CFI				BIT16			// Concatenation format indicator (CFI) bit (only valid if bit 21 is set)
#define RXBUF_STAT_EFRAME			BIT15			// End of Frame - when set the buffer contains the end of a frame
#define RXBUF_STAT_SFRAME			BIT14			// Start of Frame - when set the buffer contains the start of a frame
#define RXBUF_STAT_OFFSET0			(0x0 << 12)		// Receive buffer offset - offset = 0
#define RXBUF_STAT_OFFSET1			(0x1 << 12)		// Receive buffer offset - offset = 1 byte
#define RXBUF_STAT_OFFSET2			(0x2 << 12)		// Receive buffer offset - offset = 2 bytes
#define RXBUF_STAT_OFFSET3			(0x3 << 12)		// Receive buffer offset - offset = 3 bytes
#define RXBUF_STAT_LEN_MASK			0x7ff			// Length of frame including FCS

// Transmit Buffer Descriptor Header
typedef struct
{
	ulong address;		// transmit buffer address
	ulong status;		// status
} TXBUF_HDR;

// Bit assignments for Transmit Buffer Descriptor
// Address - Word 0
#define TXBUF_ADD_BASE_MASK			0xffffffff	// Base address of the transmit buffer in memory

// Status - Word 1
#define TXBUF_STAT_USED				BIT31		// buffer used when set
#define TXBUF_STAT_WRAP				BIT30		// set for last descriptor in list
#define TXBUF_STAT_RETRY			BIT29		// Retry limit error when set
#define TXBUF_STAT_UND				BIT28		// Transmit underrun error
#define TXBUF_STAT_BEX				BIT27		// Buffer Exhausted mid frame
#define TXBUF_STAT_NOCRC			BIT16		// Set to one to tell MAC not to add CRC to frame
#define TXBUF_STAT_LAST				BIT15		// Set to one to tell MAC we are pointing to the last buffer
#define TXBUF_STAT_LEN_MASK			0x7ff		// Length of frame including FCS

// Bit defines for the PHY Status Register #1 (phy address 0x01)
#define PHY_STAT_100BASE_T4		BIT15		// 1 = PHY able to perform 100BASE-T4
#define PHY_STAT_100BASE_X_FDX	BIT14		// 1 = PHY able to perform full-duplex 100BASE-X
#define PHY_STAT_100BASE_X_HDX	BIT13		// 1 = PHY able to perform half-duplex 100BASE-X
#define PHY_STAT_10BASE_FDX		BIT12		// 1 = PHY able to operate at 10 Mbps in full-duplex mode
#define PHY_STAT_10BASE_HDX		BIT11		// 1 = PHY able to operate at 10 Mbps in half-duplex mode
#define PHY_STAT_MF_PREAMBLE	BIT6		// 1 = PHY will accept management frames with preamble suppressed
#define PHY_STAT_AUTO_NEG_DONE	BIT5		// 1 = Auto-negotiation complete
#define PHY_STAT_REM_FLT		BIT4		// 1 = Remote fault condition detected
#define PHY_STAT_AUTO_NEG_ABLE	BIT3		// 1 = PHY is able to perform Auto-Negotiation
#define PHY_STAT_LINK_UP		BIT2		// 1 = Link is up
#define PHY_STAT_JABBER			BIT1		// 1 = Jabber condition detected
#define PHY_STAT_EXT_REG		BIT0		// 1 = Extended register capabilities


#define RX_BUFFER_SIZE 128
#define NUM_RXBDS	   32
#define TX_BUFFER_SIZE 1600


// use internal SRAM for buffers and descriptors
// this also insures that the receive descriptors
// start on a 64byte boundary
RXBUF_HDR *rxbuf_hdrs = (RXBUF_HDR *)SRAM_BASE;	// strt of receive buffer descriptors
TXBUF_HDR *txbuf_hdr = (TXBUF_HDR *)(SRAM_BASE + 0x100);	// start of transmit buffer descriptor - only one
uchar *txbuf = (uchar *)(SRAM_BASE + 0x200);		// one transmit buffer, 1536 bytes maximum
uchar *rxbuf = (uchar *)(SRAM_BASE + 0x800);		// receive buffers starting address

uchar umon_rxbuf[8][1600];
int rxidx;

#define EMAC_DBG	   0

// ********************************************
// phyread():
// Read the PHY
ulong phyread(uchar reg)
{
	EMAC_REG(EMAC_MAN) = (0x01 << 30	// Start of Frame Delimiter
					    | 0x02 << 28	// Operation, 0x01 = Write, 0x02 = Read
						| 0x1f << 23	// Phy Number, 0 as we only have one
						| reg << 18		// Phy Register
						| 0x02 << 16);	// must be 0x02 for turn around field

	// wait for phy read to complete
	udelay(1000);

	if (EMAC_DBG) {
		printf("EMAC: Phy, Reg %d, Read 0x%04lx.\n", reg, (EMAC_REG(EMAC_MAN) & 0xffff));
	}

	return EMAC_REG(EMAC_MAN) & 0xffff;
}
// ********************************************
// phywrite()
// write the PHY
void phywrite(uchar reg, ushort data)
{
	EMAC_REG(EMAC_MAN) = (0x01 << 30		// Start of Frame Delimiter
	    			   | 0x01 << 28			// Operation, 0x01 = Write, 0x02 = Read
					   | 0x1f << 23			// Phy Number, LAN8700 is address 0x1f
					   | reg << 18			// Phy Register
					   | 0x02 << 16			// must be 0x02 for turn around field
					   | data);
	if (EMAC_DBG) {
		printf("EMAC: Phy, Reg %d, Write 0x%04x.\n", reg, data);
	}

	// wait for phy write to complete
	udelay(1000);
}

/*
 * enreset():
 *	Reset the PHY and MAC.
 */
void
enreset(void)
{
	// Disable TX/RX 
	EMAC_REG(EMAC_CTL)  &= ~(EMAC_CTL_TE | EMAC_CTL_RE);
}

/*
 * eninit():
 * This would include establishing buffer descriptor tables and
 * all the support code that will be used by the ethernet device.
 *
 * It can be assumed at this point that the array uchar BinEnetAddr[6]
 * contains the 6-byte MAC address.
 *
 * Return 0 if successful; else -1.
 */
int
eninit(void)
{
	volatile int i, timeout;
	ulong emac_link_status;

	// Set RMII mode and Clock enable in User Input/Output Register
	EMAC_REG(EMAC_USRIO) =	(EMAC_USRIO_RMII | EMAC_USRIO_CLKEN);

	if (EMAC_DBG & 4) {
		printf("EMAC: RX Buffer Header Descriptor Address = 0x%08lx.\n", (ulong)rxbuf_hdrs);
	}

	// initialize our transmit buffer descriptor address, status gets
	// dealt with in the transmit section
	if (EMAC_DBG & 4) {
		printf("EMAC: TX Buffer Header Descriptor Address = 0x%08lx.\n", (ulong)txbuf_hdr);
		printf("EMAC: TX Buffer Address = 0x%08lx.\n", (ulong)txbuf);
	}
	txbuf_hdr[0].address = (ulong)(&txbuf[0]);

	// initialize our receive buffer descriptors
	for (i = 0; i < NUM_RXBDS; i++)
	{
		if (EMAC_DBG & 4) {
			printf("EMAC: RX Buffer %d Address = 0x%08lx.\n", i, (ulong)(&rxbuf[i * RX_BUFFER_SIZE]));
		}
		rxbuf_hdrs[i].address = (ulong)(&rxbuf[i * RX_BUFFER_SIZE]);
		rxbuf_hdrs[i].status = 0x00000000;
	}
	i--;
	
	// last one needs the wrap bit set
	rxbuf_hdrs[i].address = (ulong)(&rxbuf[i * RX_BUFFER_SIZE]) | RXBUF_ADD_WRAP;
	rxbuf_hdrs[i].status = 0x00000000;

	if (EMAC_DBG & 4) {
		printf("EMAC: RX Buffer %d (Last) Address = 0x%08lx.\n", i, (ulong)(&rxbuf[i * RX_BUFFER_SIZE]));
	}

	
	if (EMAC_DBG & 4) {
		printf("EMAC: RX Buffer Header Descriptor Address = 0x%08lx.\n", (ulong)rxbuf_hdrs);
	}

	// set the MAC address - must write the low first
	EMAC_REG(EMAC_SA1L) = (BinEnetAddr[3] << 24) | (BinEnetAddr[2] << 16) | (BinEnetAddr[1] << 8) | BinEnetAddr[0];
	EMAC_REG(EMAC_SA1H) = (BinEnetAddr[5] << 8) | BinEnetAddr[4];

	// point to our receive buffer queue
	EMAC_REG(EMAC_RBQP) = (ulong)rxbuf_hdrs;

	if (EMAC_DBG & 4) {
		printf("EMAC: RX Buffer Queue = 0x%08lx.\n", (ulong)rxbuf_hdrs);
	}
	// clear any left over status bits
	EMAC_REG(EMAC_RSR)  &= ~(EMAC_RSR_OVR | EMAC_RSR_REC | EMAC_RSR_BNA);

	// set the MII clock divder to MCK/64 and set copy all frames
	EMAC_REG(EMAC_CFG) = (EMAC_CFG_CLK_64 | EMAC_CFG_CAF | EMAC_CFG_BIG | EMAC_CFG_FD);

	// enable the MII interface via MDC and MDIO
	EMAC_REG(EMAC_CTL) = (EMAC_CTL_MPE);

	if (EMAC_DBG) {
		printf("EMAC: Getting Link Status.\n");
	}
	// read the PHY ID registers
	emac_link_status = phyread(0x02);
	emac_link_status = phyread(0x03);

	// Get the link status - wait for done with a timeout
	timeout = 5000;
	printf("EMAC: Auto-Negotiate...");
	while (timeout)
	{
		emac_link_status = phyread(0x01);
		if (!(emac_link_status & PHY_STAT_AUTO_NEG_ABLE)){
			printf("\b\b\b PHY is unable to Auto-Negotatiate!\n");
			timeout = 0;
			break;
		}
		if (emac_link_status & PHY_STAT_AUTO_NEG_DONE)
		{
			printf("\b\b\b complete, Link = ");
			break;
		}
		timeout-- ;
	}
	if (!(timeout))
	{
		printf("\b\b\b failed, Status = 0x%04lx!\n", emac_link_status);
		printf("EMAC: Initialization Halted.\n");
		return -1; 
		// if autonegotiation fails, just force to 10HD...
		emac_link_status = PHY_STAT_10BASE_HDX;
	}

	// Set SPD and FD based on the return link status
	if (emac_link_status & (PHY_STAT_100BASE_X_FDX | PHY_STAT_100BASE_X_HDX))
	{
		EMAC_REG(EMAC_CFG) |= EMAC_CFG_SPD;
		printf("100MBIT, ");
	}
	else
	{
		EMAC_REG(EMAC_CFG) &= ~EMAC_CFG_SPD;
		printf("10MBIT, ");
	}		
			
	if (emac_link_status & (PHY_STAT_100BASE_X_FDX | PHY_STAT_10BASE_FDX))
	{
		EMAC_REG(EMAC_CFG) |= EMAC_CFG_FD;
		printf("Full Duplex.\n");
	}
	else
	{
		EMAC_REG(EMAC_CFG) &= ~EMAC_CFG_FD;
		printf("Half Duplex.\n");
	}		

	if (EMAC_DBG) {
		printf("EMAC: Enable TX and RX.\n");
	}
	// Enable TX/RX and clear the statistics counters
	EMAC_REG(EMAC_CTL) |= (EMAC_CTL_TE | EMAC_CTL_RE | EMAC_CTL_CSR);

	return (0);

}

int
EtherdevStartup(int verbose)
{

	/* Put ethernet controller in reset: */
	enreset();

	/* Initialize controller: */
	eninit();

	return(0);
}

/*
 * polletherdev():
 * Called continuously by the monitor (ethernet.c) to determine if there
 * are any incoming ethernet packets.
 */
int
polletherdev(void)
{
	static int idx;
	char *uptr, *ubase;
	int	pktlen = 0, pktcnt = 0;
	int j, x, y, first_i, last_i;

	x = y = -1;
	// clear the receive bit as we do not use it anyway
	EMAC_REG(EMAC_RSR) = (EMAC_RSR_REC | EMAC_RSR_OVR | EMAC_RSR_BNA);

	first_i = idx;
	last_i = NUM_RXBDS;

	// scan the buffer descriptors looking for any that have data in them

loop:
	for(idx=first_i; idx < last_i; idx++)
	{
		// find the start of a frame
		if (rxbuf_hdrs[idx].address & RXBUF_ADD_OWNED)
		{
			if ((x == -1) && (rxbuf_hdrs[idx].status & RXBUF_STAT_SFRAME))
			{
				x = idx;	// remember the start of frame buffer
				if (EMAC_DBG & 2)
					printf("EMAC: SOF = %d\n", x);
				
				// We've found the start of the frame.  Now, before we go
				// any farther, make sure the end of frame is buffered up.
				// If it isn't, just return here...
				for (j=idx;j<NUM_RXBDS;j++) {
					if ((rxbuf_hdrs[j].address & RXBUF_ADD_OWNED) == 0) {
						if (EMAC_DBG & 2)
							printf("Incomplete pkt_1\n");
						return(0);
					}
					if (rxbuf_hdrs[j].status & RXBUF_STAT_EFRAME)
						break;
				}
				if (j == NUM_RXBDS) {
					for(j=0;j<idx;j++) {
						if ((rxbuf_hdrs[j].address & RXBUF_ADD_OWNED) == 0) {
							if (EMAC_DBG & 2)
								printf("Incomplete pkt_2\n");
							return(0);
						}
						if (rxbuf_hdrs[j].status & RXBUF_STAT_EFRAME)
							break;						
					}
					if (j == idx) {
						if (EMAC_DBG & 2)
							printf("Incomplete pkt_3\n");
						return(0);
					}
				}
			}
				
			// After we find the SOF (x != -1), then we look for the EOF.			
			if ((x != -1) && (rxbuf_hdrs[idx].status & RXBUF_STAT_EFRAME))
			{
				pktlen = rxbuf_hdrs[idx].status & RXBUF_STAT_LEN_MASK;
				y = idx;	// remember the end of frame buffer

				if (EMAC_DBG & 2) {
					printf("EMAC: EOF = %d\n", y);

					printf("EMAC: RX Buffer %d Address = 0x%08lx, Status = 0x%08lx.\n", 
							idx, rxbuf_hdrs[x].address, rxbuf_hdrs[y].status);
					printf("EMAC: Releasing buffers %d to %d\n", x, y);
				}
				pktcnt++;
				EtherRFRAMECnt++;
				uptr = ubase = (char *)umon_rxbuf[rxidx];
				rxidx++;
				if (rxidx == 8) rxidx = 0;
				
				// Copy the data to the umon buffer and release the buffers back to the EMAC
				for (j = x;; j++)
				{
					if (j == NUM_RXBDS)
						j = 0;
					rxbuf_hdrs[j].address &= ~RXBUF_ADD_OWNED;
					memcpy(uptr,(char *)(rxbuf_hdrs[j].address & 0xfffffffc),128);
					if (j == y)
						break;
					uptr += 128;
				}
				processPACKET((struct ether_header *)ubase, pktlen);
				if (EMAC_DBG & 2)
					printf("EMAC: Process Packet Complete, Length = 0x%08lx, Count = %d.\n", pktlen, pktcnt);

				return(pktcnt);
			}
		}
	}
	if (first_i != 0) {
		last_i = first_i; 
		first_i = 0;
		goto loop;
	}

	// make sure the receiver/transmitter are still enabled
//	EMAC_REG(EMAC_CTL) |= (EMAC_CTL_RE | EMAC_CTL_TE);

	return(pktcnt);
}


/* getXmitBuffer():
 * Return a pointer to the buffer that is to be used for transmission of
 * the next packet.  Since the monitor's driver is EXTREMELY basic,
 * there will only be one packet ever being transmitted.  No need to queue
 * up transmit packets.  Since the SAM89263 uses buffer descriptors for
 * transmit, this code would need significant changes to support mutliple 
 * buffers.  For now just use one buffer.
 */
uchar *
getXmitBuffer(void)
{
	return txbuf;
}

/* sendBuffer():
 * Send out the packet assumed to be built in the buffer returned by the
 * previous call to getXmitBuffer() above.
 */
int
sendBuffer(int length)
{
	ulong status32;
	int timeout;

	if (EMAC_DBG & 4) {
		printf("EMAC: TX Buffer Address = 0x%08lx, Length = %d.\n", (ulong)txbuf, length);
		printf("EMAC: TX Buffer Header is at 0x%08lx.\n", (ulong)txbuf_hdr);
	}
#if INCLUDE_ETHERVERBOSE
	if (EtherVerbose &  SHOW_OUTGOING)
		printPkt((struct ether_header *)txbuf,length,ETHER_OUTGOING);
#endif

	/* Bump up the packet length to a minimum of 64 bytes.
	 */
	if (length < 64)
		length = 64;

	// clear any pending status
	EMAC_REG(EMAC_TSR) = (EMAC_TSR_COL | EMAC_TSR_RLE | EMAC_TSR_COMP  
					   | EMAC_TSR_UND | EMAC_TSR_BEX); 

	// tell the EMAC about our buffer
	// point to our transmit buffer queue
	txbuf_hdr[0].status = (ulong)length | TXBUF_STAT_WRAP | TXBUF_STAT_LAST;
	EMAC_REG(EMAC_TBQP) = (ulong)txbuf_hdr;
	EtherXFRAMECnt++;

	if (EMAC_DBG & 4) {
	printf("EMAC: TX Buffer Status = 0x%08lx \n", (ulong)txbuf_hdr[0].status);
	}

	// Start Transmitting
	EMAC_REG(EMAC_CTL) |= (EMAC_CTL_TST);

	// wait until tx is complete.  This acts as a throttle to prevent
	// receiver overruns
	timeout = 10000;
	while(timeout){
		if (EMAC_REG(EMAC_TSR) & EMAC_TSR_COMP) break;
		timeout--;
	}

	// clean up descriptor status
	txbuf_hdr[0].status = 0;

	// check if there was an error
	if (EMAC_DBG & 4) {
		status32 = EMAC_REG(EMAC_TSR);
		if (status32 & (EMAC_TSR_COL | EMAC_TSR_RLE | EMAC_TSR_UND | EMAC_TSR_BEX)){
			printf("EMAC: TX Error Status = 0x%08lx.\n", status32);
		}
	}

	return 0;
}

/* DisableEtherdev():
 * Fine as it is...
 */
void
DisableEtherdev(void)
{
	enreset();
}

/* ShowEtherdevStats():
 * This function is used to display device-specific stats (error counts
 * usually).
 */

void
ShowEtherdevStats()
{
	int i, phydata;

	printf("EMAC Transmit Frames OK = %d\n",(int)EMAC_REG(EMAC_FRA));
	printf("EMAC Single Collision Frames = %d\n",(int)EMAC_REG(EMAC_SCOL));
	printf("EMAC Multiple Collision Frames = %d\n",(int)EMAC_REG(EMAC_MCOL));
	printf("EMAC Received Frames OK = %d\n",(int)EMAC_REG(EMAC_OK));
	printf("EMAC SEQE Errors = %d\n",(int)EMAC_REG(EMAC_SEQE));
	printf("EMAC Alignment Errors = %d\n",(int)EMAC_REG(EMAC_ALE));
	printf("EMAC Deffered Transmit Frames = %d\n",(int)EMAC_REG(EMAC_DTE));
	printf("EMAC Late Collision= %d\n",(int)EMAC_REG(EMAC_LCOL));
	printf("EMAC Excessive Collision Errors = %d\n",(int)EMAC_REG(EMAC_ECOL));
	printf("EMAC Carrier Sense Errors = %d\n",(int)EMAC_REG(EMAC_CSE));
	printf("EMAC Transmit Underrun Errors = %d\n",(int)EMAC_REG(EMAC_TUE));
	printf("EMAC Receive Buffer Not Available Errors = %d\n",(int)EMAC_REG(EMAC_RRE));
	printf("EMAC Receive DMA Overrun Errors = %d\n",(int)EMAC_REG(EMAC_ROV));
	printf("EMAC Receive Length Mismatch Errors = %d\n",(int)EMAC_REG(EMAC_SEQE));
	printf("EMAC Receive Frame Too Big Errors = %d\n",(int)EMAC_REG(EMAC_ELR));
	printf("EMAC Receive Jabber Errors = %d\n",(int)EMAC_REG(EMAC_RJB));
	printf("EMAC Receive Frame Too Small Errors = %d\n",(int)EMAC_REG(EMAC_USF));


//    phywrite(0x1f, 0); 	// Clear shadow bit
	
    for(i = 0; i < 0x20; i++)
	{
		phydata = phyread(i);
        printf("PHY REG %02x = 0x%02x\n" ,i, phydata);
	}
    // BCM5221 has some shadow registers we are interested in as well

//	phywrite(0x1f, BIT7); 	// Set Bit 7 - Shadow register enable
//    for(i = 0; i < 36; i++)
//	{
//		phydata = phyread(i);
//        printf("PHY Shadow REG %02x = 0x%02x\n" ,i, phydata);
//	}
}

/* extGetIpAdd():
 * If there was some external mechanism (other than just using the
 * IPADD shell variable established in the monrc file) for retrieval of
 * the board's IP address, then do it here...
 */
char *
extGetIpAdd(void)
{
	return((char *)0);
}

/* extGetEtherAdd():
 * If there was some external mechanism (other than just using the
 * ETHERADD shell variable established in the monrc file) for retrieval of
 * the board's MAC address, then do it here...
 */
char *
extGetEtherAdd(void)
{
	return((char *)0);
}

/* disablePromiscuousReception():
 * Provide the code that disables promiscuous reception.
 */
void
disablePromiscuousReception(void)
{
	/* OPT_ADD_CODE_HERE */	 
}

/* enablePromiscuousReception():
 * Provide the code that enables promiscuous reception.
 */
void
enablePromiscuousReception(void)
{
	/* OPT_ADD_CODE_HERE */
}

/* disableBroadcastReception():
 * Provide the code that disables broadcast reception.
 */
void
disableBroadcastReception(void)
{
	/* ADD_CODE_HERE */
}

/* enableBroadcastReception():
 * Provide the code that enables broadcast reception.
 */
void
enableBroadcastReception(void)
{
	/* ADD_CODE_HERE */
}

/* 
 * enselftest():
 *	Run a self test of the ethernet device(s).  This can be stubbed
 *	with a return(1).
 *	Return 1 if success; else -1 if failure.
 */
int
enselftest(int verbose)
{
	return(1);
}

#endif
