/* cpuio.c:

	General notice:
	This code is part of a boot-monitor package developed as a generic base
	platform for embedded system designs.  As such, it is likely to be
	distributed to various projects beyond the control of the original
	author.  Please notify the author of any enhancements made or bugs found
	so that all may benefit from the changes.  In addition, notification back
	to the author will allow the new user to pick up changes that may have
	been made by other users after this version of the code was distributed.

	Author:	Ed Sutter
	email:	esutter@lucent.com
	phone:	908-582-2351
*/
#include "config.h"
#include "cpu.h"
#include "cpuio.h"
#include "genlib.h"
#include "ether.h"
#include "stddefs.h"

/* InitUART():
 * This function is called at startup, application exit, or by the 
 * mstat -b BAUD command.
 * For startup and mstat, the baud rate will be specified and used; for
 * application exit, it will be 0 to indicate that whatever baud rate was
 * used prior by mstat or startup should be used again.
 */
void
InitUART(int baud)
{
	int i, nn, brr;

	/* Disable the transmitter and receiver and all SCI interrupts */
    /* clear TE and RE bits and select the clock source */
    *SCR0 = 0x00;

	if (baud == DEFAULT_BAUD_RATE) {
		*BRR0 = SCI_BRR_VALUE;
		*SMR0 = 0+SCI_SCR_N_VALUE;
	}
	else {
	    /* Select Bit rate counter */
		for (nn=0;nn<4;nn++) {
			brr = ((2*MON_CPU_CLOCK/32/(1<<(2*nn))/baud+1)/2-1);
			if (brr  < 256) {
				*BRR0 = (uchar)brr;
				/* Select the communications format */
			    /* xxxxxxnn internal system clock phi from 300cpu.h */
			    /* xxxxx0xx disable the multiprocessor function */
			    /* xxxx0xxx stop bit length=> 1 bit */
			    /* xxx0xxxx even parity */
			    /* xx0xxxxx no parity bit is added */
			    /* x0xxxxxx 8 bit character length in asynchrounous mode */
			    /* 0xxxxxxx asynchrounous mode */
		    	*SMR0 = 0+nn;
				break;
			}
		}
		if (nn == 4) {
			*BRR0 = SCI_BRR_VALUE;
			*SMR0 = 0+SCI_SCR_N_VALUE;
		}
	}

	for (i=0;i<100000;i++);	/* wait >= 1 bit time */

	/* Enable transmitter and receiver and receive interrupts */
    /* xxxxxx00 CKE0 internal clock, SCK pin available for generic i/o */
    /* xxxxx0xx CKE1 disables transmit end interrupts */
    /* xxxx0xxx MPIE disables multiprocessor interrupts */
    /* xxx1xxxx RE enable the receiver */
    /* xx1xxxxx TE enable the transmitter */
    /* x0xxxxxx RIE enable receive-data-full interrupts and receive-error int */
    /* 0xxxxxxx TIE enable disables transmit-data-empty interrupts */
    *SCR0 = (SCI_RE | SCI_TE);
}

int
ConsoleBaudSet(int baud)
{
	InitUART(baud);
	return(0);
}

/* The common version of start.c calls devInit(baud) instead of the older 
 * InitUART(baud).  To become compatible with this, we encapsulate InitUART()
 * with devInit() here... 
 */
int
devInit(int baud)
{
	InitUART(baud);
	return(0);
}

void
initCPUio()
{
	/* Initialize the interrupt control registers: */
	*IPRA = 0;
	*IPRB = 0;
	*IPRC = 0;
	*IPRD = 0;
	*IPRE = 0;
	*IPRF = 0;
	*IPRG = 0;
	*IPRH = 0;
	*ICR = 0;
	*ISR = 0;
}

int
target_console_empty(void)
{
	/* Wait for empty tx buffer.
	 */
	if (*SSR0 & SCI_TDRE)
		return(1);
	return(0);
}

int
target_putchar(char c)
{
	uchar	ssr0;

	/* Wait for empty tx buffer. */
	while(1) {
		ssr0 = *SSR0;
		if (ssr0 & SCI_TDRE)
	 		break;
	}

	/* write data into TDR */
	*TDR0 = c;

	/* Read and clear TDRE in SSR */
	*SSR0 &= ~SCI_TDRE;

	return((int)c);
}

int
target_getchar(void)
{
	char	ch;

	/* read char */
	ch = *RDR0;

	/* clear rx interrupt to allow next char */
	*SSR0 &= ~SCI_RDRF;

	return((int)ch);
}

int
target_gotachar(void)
{
	if(*SSR0 & SCI_RDRF)
		return(1);
	else
		return(0);
}

void
target_reset(void)
{
	coldstart();
}

void
ledsOn()
{
	*PEDR &= ~0xD800;
}

void
ledsOff()
{
	*PEDR |= 0xD800;
	*PEDR &= ~0x2000;	/* Enable LAN activity LED */
}

void
ledBlink(int count)
{
	int	i;

	while(count) {
		for(i=0;i<50000;i++);
		ledsOn();
		for(i=0;i<50000;i++);
		ledsOff();
		count--;
	}
}

void
putvbr(ulong vbr)
{
	__asm__("mov.l @r14,r0");
	__asm__("ldc r0,vbr");
}

void
putpr(ulong pr)
{
	__asm__("mov.l @r14,r0");
	__asm__("lds r0,pr");
}

void
putsr(ulong sr)
{
	__asm__("mov.l @r14,r0");
	__asm__("ldc r0,sr");
}

int
getvbr(void)
{
	__asm__("stc vbr,r0");
}

int
getgbr(void)
{
	__asm__("stc gbr,r0");
}

int
getsr(void)
{
	__asm__("stc sr,r0");
}

int
getsp(void)
{
	__asm__("mov r15,r0");
}

void
intsrestore(unsigned long sr)
{
	putsr(sr);
}

ulong
intsoff()
{
	ulong  sr;

	sr = getsr();
	putsr(sr | 0x00f0);
	return(sr);
}

#if 0
void
prHex(int size,uchar *cp,char *prefix, char *suffix)
{
	uchar	*end;
	static	char *hexdigits = "0123456789abcdef";

	if (prefix) {
		while(*prefix)
			putchar(*prefix++);
	}
	end = cp + size;
	while(cp < end) {
		putchar(hexdigits[(int)((*cp&0xf0)>>4)]);
		putchar(hexdigits[(int)(*cp&0x0f)]);
		cp++;
	}
	if (suffix) {
		while(*suffix)
			putchar(*suffix++);
	}
}
#endif
