/* config.h:
 *	Monitor configuration file for Bosch MPR2 (SH7720 based) board.
 *	
 *	General notice:
 *	This code is part of a boot-monitor package developed as a generic base
 *	platform for embedded system designs.  As such, it is likely to be
 *	distributed to various projects beyond the control of the original
 *	author.  Please notify the author of any enhancements made or bugs found
 *	so that all may benefit from the changes.  In addition, notification back
 *	to the author will allow the new user to pick up changes that may have
 *	been made by other users after this version of the code was distributed.
 *
 *	Author:	Torsten Tetz, Bosch Security Systems
 *	        Mark Jonas, Bosch Security Systems, mark.jonas@de.bosch.com
 *	        Joerg Tuermer, Bosch Security Systems, joerg.tuermer@de.bosch.com
 *	based on template/config.h file by Ed Sutter
 */
#ifndef __CONFIG_H
#define __CONFIG_H

/* Specify CPU/PLATFORM type and name so that common code can be used
 * for a similar cpu, on different platforms.
 * The 'TYPE' definition is used for ifdefs in the code and the 'NAME'
 * is used for printfs in the code.
 */
#define CPU_NAME	"Renesas SH7720 (SH3-DSP core)"
#define PLATFORM_NAME	"Bosch MPR2"

/* Set Pannel Release:
 * Allowed values are 2 or 3.
 */
#define PLATFORM_MP_R2	2

/* Define chip frequenzy in Hz.
 * Allowed values are 11000000, 12000000 or 16500000.
 */
#define QUARTZ	12000000

/* Define the frequenzy of the peripheral clock.
 */
#define PERICLCKFREQ	24000000

/* SCIF0 settings.
 * Set the default baud rate for the serial port monitor.
 */
#define DEFAULT_BAUD_RATE	115200

/* Approximately the size of a loop that will cause a 1-second delay.
 * This can be guestimated or modified with the sleep -c command at the
 * monitor command line.  
 */
#define LOOPS_PER_SECOND	4200000

/* Ticks of the board based timer (TMU) in one millisecond.
 * If the board based timer is used "#define LOOPS_PER_SECOND" is
 * not needed.
 */
#define TIMER_TICKS_PER_MSEC	((PERICLCKFREQ / 4)/1000000)*1000

/* Specify the size of the memory block (in monitor space) that is to be
 * allocated to malloc in the monitor.  Note that this size can be dynamically
 * increased using the heap extension option in the heap command.
 */
#define ALLOCSIZE	64*1024

/* Specify the size of the stack (inside of the monitor space) that is to be
 * used by the monitor.
 */
#define MONSTACKSIZE 	0x2000

/* LAN definitions
 * DRIVER_SMSC911X_BASE is the memory offset needed to access the lan
 * controller.
 */
#define DRIVER_SMSC911X_BASE	0xA8000000

/* Establish default MAC/IP addresses...
 * For the first 3 bytes of the MAC address use "ac:de:48:", 
 * which is for development, and won't be used on commercially
 * available cards.
 */
#define DEFAULT_ETHERADD	"ac:de:48:20:14:01"
#define DEFAULT_IPADD		"192.168.0.2"

/* Flash bank configuration: */
#define SINGLE_FLASH_DEVICE	1
#define FLASHBANKS		1
#define FLASH_BANK0_WIDTH	2
#define FLASH_BANK0_BASE_ADDR	0xA0000000
#define FLASH_COPY_TO_RAM
/* Macro to use the uncached P2 Area when editing "etheraddr" or writing to flash. */
#define TO_FLASH_ADDR(etherAddr)	(((unsigned long)etherAddr & 0x0FFFFFFF) | 0xA0000000)

/* TFS definitions:
 * TFSSTART:		Base address in FLASH at which TFS starts.
 * TFSEND:		End address of TFS in FLASH.
 * TFSSPARE:		Location of sector that is used as the spare sector
 * 			by TFS for defragmentation.
 * TFSSPARESIZE:	Size of the spare sector used by TFS for defragmentation.
 * TFSSECTORCOUNT:	Number of eraseable sectors that TFS covers, not including
 * 			the TFSSPARE sector.
 */
#define TFSSTART	0xA0040000	/* Sector 2 - first two sectors reserved for uMon. */
#define TFSEND		0xA0FDFFFF	/* End of sector 126 */
#define TFSSPARE	0xA0FE0000	/* Sector 127 - second half of Flash reserved eg. for Linux. */
#define TFSSPARESIZE	0x20000		/* 128 kB */
#define TFSSECTORCOUNT	125		/* Sector 2 - 126 */
#define TFS_EBIN_ELF	1
#define TFS_DISABLE_AUTODEFRAG

/* Width macros used by TFS and FLASH operations.
 * Note that only one of WIDTH8, WIDTH16 and WIDTH32 should be set and the
 * WIDTHXX_NOTUESD macros should reflect the opposite state of the WIDTH
 * macros.
 */
#define WIDTH8		0
#define WIDTH16		1
#define WIDTH32		0
#define WIDTH8_NOTUSED	1
#define WIDTH16_NOTUSED	0
#define WIDTH32_NOTUSED	1

/* FLASH_PROTECT_RANGE:
 * These sectors are configured to be software-protected by the monitor's
 * flash command.  Note this is a "software" protection mechanism, not
 * hardware!
 */
#define FLASH_PROTECT_RANGE	"0-1"	/* First two sectors reserved for uMon. */

/* Use Little_Endian CPU. */
#define CPU_LE

/* Choose the abilities to be included into the monitor. */
#define FORCE_BSS_INIT			0

#define INCLUDE_MEMCMDS			1
#define INCLUDE_EDIT			0
#define INCLUDE_DISASSEMBLER		0
#define INCLUDE_UNZIP			0
#define INCLUDE_ETHERNET		1
#define INCLUDE_ETHERVERBOSE		1
#define INCLUDE_ICMP			1
#define INCLUDE_TFTP			1
#define INCLUDE_DHCPBOOT		1
#define INCLUDE_TFS			1
#define INCLUDE_TFSAPI			1
#define INCLUDE_TFSSYMTBL		1
#define INCLUDE_TFSSCRIPT		1
#define INCLUDE_TFSCLI			1
#define INCLUDE_XMODEM			1
#define INCLUDE_LINEEDIT		1
#define INCLUDE_EE			0
#define INCLUDE_FLASH			1
#define INCLUDE_GDB			0
#define INCLUDE_STRACE			0
#define INCLUDE_CAST			0
#define INCLUDE_STRUCT			0
#define INCLUDE_MALLOC			1
#define INCLUDE_SHELLVARS		1
#define INCLUDE_REDIRECT		0
#define INCLUDE_QUICKMEMCPY		0
#define INCLUDE_PROFILER		0
#define INCLUDE_BBC			0
#define INCLUDE_MEMTRACE		1
#define INCLUDE_STOREMAC		1
#define INCLUDE_VERBOSEHELP		1
#define INCLUDE_HWTMR			1
#define INCLUDE_PORTCMD			0
#define INCLUDE_USRLVL			0

/* This is an include file used to verify that all of the monitor macros
 * are defined. There is also some attempt to verify that macro conflicts
 * do not exist (for example, the flash file system cannot exist if the
 * flash driver is not included); however, there are probably holes in this
 * check.
 */
#include "inc_check.h"

#endif	/* __CONFIG_H */

