/* smsc911x.c:
 *	This file contains the ethernet driver for the SMSC LAN911x,
 *	for use with MicroMonitor v1.11.
 *	The current version was only tested with the LAN9115 but supports
 *	other LAN911x and LAN921x family members as well.
 *
 *	General notice:
 *	This code is part of a boot-monitor package developed as a generic base
 *	platform for embedded system designs.  As such, it is likely to be
 *	distributed to various projects beyond the control of the original
 *	author.  Please notify the author of any enhancements made or bugs found
 *	so that all may benefit from the changes.  In addition, notification back
 *	to the author will allow the new user to pick up changes that may have
 *	been made by other users after this version of the code was distributed.
 *
 *	Author:	Torsten Tetz, Bosch Security Systems
 */
#include "config.h"
#include "stddefs.h"
#include "genlib.h"
#include "monlib.h"
#include "timer.h"
#include "ether.h"
#include "smsc911x.h"

/* smsc911x_start_elapsed_timer():
 * Start the timer from given time. Timer does a count down.
 * The function makes use of MicroMonitor's integrated timer.
 * When MicroMonitor is set to use a board based timer, 
 * this timer is rather accurate.
 */
void
smsc911x_start_elapsed_timer(struct elapsed_tmr *tmr,int milliseconds)
{
	startElapsedTimer(tmr,milliseconds);
}

/* smsc911x_msec_elapsed():
 * Return 0 when time set is not elapsed yet.
 * Return 1 when time set elapsed.
 * Function to make use of MicroMonitor's integrated timer.
 * When MicroMonitor is set to use a board based timer, 
 * this timer is rather accurate.
 */
int
smsc911x_msec_elapsed(struct elapsed_tmr *tmr)
{
	return(msecElapsed(tmr));
}

/* smsc911x_verify_chip_ID():
 * Return 0 when chip ID was found and is indicating a known chip.
 * Return -1 if no chip ID could be found, or found ID is not
 * indicating a known chip..
 */
int
smsc911x_verify_chip_ID(void)
{
	unsigned long idAndRevision = 0;

	/* After Power on, wake up or when returning from a power saving
	 * state, the LAN911x must be read at least once to make write 
	 * opperations possible.
	 */
	if(SMSC911X_GET_REG(BYTE_TEST) == 0x87654321) {

		idAndRevision = ((SMSC911X_GET_REG(ID_REV)) >> 16);
		
		if(idAndRevision == 0) {
				printf(NAME_OF_NETDEVICE": Could not get chip ID.\n");
				return(-1);
		}

		switch(idAndRevision) {	
			case 0x115:	/* Check for smsc9115. */
				if(EtherVerbose & SHOW_DRIVER_DEBUG) {
					printf(NAME_OF_NETDEVICE": Chip found: SMSC9115\n");
				}
				return(0);
			case 0x116:	/* Check for smsc9116. */
				if(EtherVerbose & SHOW_DRIVER_DEBUG) {
					printf(NAME_OF_NETDEVICE": Chip found: SMSC9116\n");
				}
				return(0);
			case 0x117:	/* Check for smsc9117. */
				if(EtherVerbose & SHOW_DRIVER_DEBUG) {
					printf(NAME_OF_NETDEVICE": Chip found: SMSC9117\n");
				}
				return(0);
			case 0x118:	/* Check for smsc9118. */
				if(EtherVerbose & SHOW_DRIVER_DEBUG) {
					printf(NAME_OF_NETDEVICE": Chip found: SMSC9118\n");
				}
				return(0);
			case 0x115A:	/* Check for smsc9215. */
				if(EtherVerbose & SHOW_DRIVER_DEBUG) {
					printf(NAME_OF_NETDEVICE": Chip found: SMSC9215\n");
				}
				return(0);
			case 0x118A:	/* Check for smsc9218. */
				if(EtherVerbose & SHOW_DRIVER_DEBUG) {
					printf(NAME_OF_NETDEVICE": Chip found: SMSC9218\n");
				}
				return(0);
			default:
				printf(NAME_OF_NETDEVICE": Unknown chip ID.\n");
				return(-1);
		}
	}
	else {
		printf(NAME_OF_NETDEVICE": Byte order test failed.\n");
	}
	return(-1);
}

/* smsc911x_set_MIIPHY_reg():
 * This function sets the value of a given PHY register, indirectly trough the MAC_CSR
 * and further indirectly trough the MII Registers.
 * The phyAddress parameter selects internal or external PHY. The internal PHY
 * is set to address 00001b.
 */
void
smsc911x_set_MIIPHY_reg(unsigned char phyAddress,unsigned char miiRegister,unsigned short data)
{
	unsigned long mii_acc_setvalue = 0;
	struct elapsed_tmr tmr;
	
	if(!(smsc911x_get_MAC_CSR(MII_ACC) & MII_ACC_MII_BUSY_)) {
		
		/* Place data for writing. */
		smsc911x_set_MAC_CSR(MII_DATA,data);
		
		/* Set where the data should be written to and set the busy bit to start. */
		mii_acc_setvalue = (phyAddress << 11) | (miiRegister << 6) | MII_ACC_MII_WRITE_ ;
		mii_acc_setvalue = mii_acc_setvalue | MII_ACC_MII_BUSY_ ;
		smsc911x_set_MAC_CSR(MII_ACC,mii_acc_setvalue);

		/* Wait for write to complete, with timeout. */
		smsc911x_start_elapsed_timer(&tmr,1000);
		do {
			if(!(smsc911x_get_MAC_CSR(MII_ACC) & MII_ACC_MII_BUSY_)) {
				return;
			}
		} while(!smsc911x_msec_elapsed(&tmr));
		printf(NAME_OF_NETDEVICE ": Writing to PHY timed out.\n");
	}
	else {
		printf(NAME_OF_NETDEVICE ": MII was busy before writing.\n");
	}
	return;
}

/* smsc911x_get_MIIPHY_reg():
 * This function fetches the value of a given PHY register, indirectly 
 * trough the MAC_CSR and further indirectly trough the MII Registers.
 * The "phyAddress" parameter selects internal or external PHY. The internal PHY
 * is set to address 00001b.
 * The function Returns the value of the given PHY register or 0 when an 
 * error has occured.
 */
unsigned short
smsc911x_get_MIIPHY_reg(unsigned char phyAddress,unsigned char miiRegister)
{
	unsigned long mii_acc_setvalue = 0;
	struct elapsed_tmr tmr;
	
	if(!(smsc911x_get_MAC_CSR(MII_ACC) & MII_ACC_MII_BUSY_)) {
		
		/* Set where the data should be read from and set the
		 *  busy bit to start.
		 */
		mii_acc_setvalue = (phyAddress << 11) | (miiRegister << 6) ;
		mii_acc_setvalue = mii_acc_setvalue | MII_ACC_MII_BUSY_ ;
		smsc911x_set_MAC_CSR(MII_ACC,mii_acc_setvalue);
		
		/* Wait for read to complete, with timeout. */
		smsc911x_start_elapsed_timer(&tmr,1000);
		do {
			if(!(smsc911x_get_MAC_CSR(MII_ACC) & MII_ACC_MII_BUSY_)) {

				/* Get the data. */
				return((unsigned short)smsc911x_get_MAC_CSR(MII_DATA));
			}
		} while(!smsc911x_msec_elapsed(&tmr));
		printf(NAME_OF_NETDEVICE ": Reading from PHY timed out.\n");
	}
	else {
		printf(NAME_OF_NETDEVICE ": MII was busy before reading.\n");
	}
	return(0);
}

/* smsc911x_MAC_BUSY():
 * Function waits for MAC_CSR_CMD_CSR_BUSY_ to become 0, with
 * a timeout.
 * Function returns 1 when MAC_CSR_CMD_CSR_BUSY_ is cleared
 * and 0 when MAC_CSR_CMD_CSR_BUSY_ is still set (after the
 * timeout).
 */
unsigned char
smsc911x_MAC_NOT_BUSY(void)
{
	struct elapsed_tmr tmr;

	/* Use 100 milliseconds as timeout. */
	smsc911x_start_elapsed_timer(&tmr,100);

	do {
		if(!(SMSC911X_GET_REG(MAC_CSR_CMD) & MAC_CSR_CMD_CSR_BUSY_)) {
			return(1);
		}
	} while(!smsc911x_msec_elapsed(&tmr));

	printf(NAME_OF_NETDEVICE ": Timeout, while waiting for MAC NOT BUSY. MAC_CSR_CMD = 0x%08lX\n", SMSC911X_GET_REG(MAC_CSR_CMD));
	return(0);
}

/* smsc911x_set_MAC_CSR():
 * This function is used by others, to set values inside of the
 * MAC Control and Status Registers indirectly trough the
 * MAC_CSR_CMD and the MAC_CSR_DATA registers.
 */
void 
smsc911x_set_MAC_CSR(unsigned char registerSymbol,unsigned long data)
{
	if(smsc911x_MAC_NOT_BUSY()) {

		/* Place data in hold register. */
		SMSC911X_SET_REG(MAC_CSR_DATA,data);

		/* Set MAC_CSR to write data to. */
		SMSC911X_SET_REG(MAC_CSR_CMD,registerSymbol);

		/* Set BUSY flag to start writing. */
		SMSC911X_SET_REG_BITS(MAC_CSR_CMD,MAC_CSR_CMD_CSR_BUSY_);

		/* Wait for wrinting to finish. */
		if(smsc911x_MAC_NOT_BUSY()) {
			return;
		}
		else {
			printf(NAME_OF_NETDEVICE ": Writing to MAC_CSR lasted to long.\n");
		}
	}
	else {
		printf(NAME_OF_NETDEVICE ": MAC_CSR was busy before writing.\n");
	}
	return;
}

/* smsc911x_get_MAC_CSR():
 * This function is used by others, to get values from the
 * MAC Control and Status Registers.
 * Returns -1 when reading fails.
 */
unsigned long
smsc911x_get_MAC_CSR(unsigned char registerSymbol)
{
	if(smsc911x_MAC_NOT_BUSY()) {

		/* Set MAC_CSR to get contents from. */
		SMSC911X_SET_REG(MAC_CSR_CMD,registerSymbol);

		/* Set R flag to indicate reading instead of wrinting. */
		SMSC911X_SET_REG_BITS(MAC_CSR_CMD,MAC_CSR_CMD_R_NOT_W_);

		/* Set BUSY flag to start reading. */
		SMSC911X_SET_REG_BITS(MAC_CSR_CMD,MAC_CSR_CMD_CSR_BUSY_);

		/* Wait for reading to finish. */
		if(smsc911x_MAC_NOT_BUSY()) {
			return(SMSC911X_GET_REG(MAC_CSR_DATA));			
		}
		else {
			printf(NAME_OF_NETDEVICE ": Reading from MAC_CSR lasted to long.\n");
		}
	}
	else {
		printf(NAME_OF_NETDEVICE ": MAC_CSR was busy before reading.\n");
	}
	return(-1);
}

/* smsc911x_reset():
 * Set Power state to Normal, initiate a soft reset.
 * (This soft reset does not reset the PHY registers.)
 */
void 
smsc911x_reset(void)	
{
	struct elapsed_tmr tmr;
	int timeout = 1;
		
	/* After Power on, wake up or when returning from a power saving
	 * state, the LAN911x must be read at least once to
	 * make write opperations possible.
	 * Make sure the SMSC911x is in D0 (Normal) state: 
	 */
	if(!(SMSC911X_GET_REG_BITS(PMT_CTRL,PMT_CTRL_READY_))) {

		/* Writing something to BYTE_TEST will wake up the device to the
		 * D0 (Normal) state, when it was in a power saving mode.
		 * Only in D0 state, write operations are allowed.
		 */
		SMSC911X_SET_REG(BYTE_TEST,0x0);

		/* Wait for READY bit to be set.
		 * Use 100 milliseconds as timeout.
		 */
		smsc911x_start_elapsed_timer(&tmr,100);
		while(!smsc911x_msec_elapsed(&tmr)) {

			/* Check if state D0 has been reached. */
			if(SMSC911X_GET_REG_BITS(PMT_CTRL, PMT_CTRL_READY_)) {
				timeout = 0;
				break;
			}
		}
				
		/* Timeout has occured, when device was still not in D0 state. */
		if(timeout) {
			printf(NAME_OF_NETDEVICE 
				": Reset timeout while waiting for normal power state.\n");
			return;
		}
	}

	/* Initiate a soft reset. 
	 * Resets all of the MAC CSR's. The SCSR's (system command and
	 * status registers) are reset except for any NASR bits.
	 * Any TX and RX errors are cleared.
	 * EEPROM reload cycle is started.
	 * The SRST (Soft Reset) bit is self-clearing.
	 */
	SMSC911X_SET_REG_BITS(HW_CFG,HW_CFG_SRST_);

	/* Wait for reset to complete.
	 * Reloading the EEPROM should complete in under 1 millisecond.
	 */
	smsc911x_start_elapsed_timer(&tmr,1);
	timeout = 1;
	while(!smsc911x_msec_elapsed(&tmr)) {

		/* Check if EEPROM reload is finised. */
		if(!(SMSC911X_GET_REG(E2P_CMD) & E2P_CMD_EPC_BUSY_)) {
			timeout = 0;
			break;
		}
	}
	if(timeout) {
		printf(NAME_OF_NETDEVICE 
			": Reset timeout while waiting for soft reset to finish.\n");
		return;
	}
return;
}

/* smsc911x_PHY_reset():
 * Soft resets the PHY registers. 
 * After a PHY reset the MAC_CR_RXEN_ and MAC_CR_TXEN_ bits in MAC_CR 
 * need to be set, in order to send and receive frames.
 * The PHY_BSR_LINK_STATUS (RO) bit must be checked before attempting to
 * receive or transmit data. (All done in "smsc911x_PHY_configure()" and 
 * "smsc911x_enable()").
 */
void
smsc911x_PHY_reset(void)
{
	struct elapsed_tmr tmr;

	SMSC911X_SET_REG_BITS(PMT_CTRL,PMT_CTRL_PHY_RST_);
	
	/* Use 1 millisecond as timeout. Should not take much more time than 100us. */
	smsc911x_start_elapsed_timer(&tmr,1);

	do {
		if(!(SMSC911X_GET_REG_BITS(PMT_CTRL,PMT_CTRL_PHY_RST_))) {
			return;
		}
	} while(!smsc911x_msec_elapsed(&tmr));

	printf(NAME_OF_NETDEVICE ": Timeout, while reseting PHY.\n");
	return;
}

/* smsc911x_PHY_configure():
 * This function is only used by "smsc911x_init()",
 * to encapsule the PHY configuration.
 * Here is only the PHY configuration, no enabling of RX and TX.
 */
void 
smsc911x_PHY_configure(void)
{
	struct elapsed_tmr tmr;
	unsigned short status = 0;
	
	/* Enable 10Mbit HD, 10Mbit FD, 100Mbit HD and 100Mbit FD abilities.
	 * Set PHY_SELECTOR_FIELD_ to IEEE 802.3 .
	 */
	smsc911x_set_MIIPHY_reg(1,PHY_ANEG_ADV, (PHY_ANEG_ADV_10H_ | PHY_ANEG_ADV_10F_ | PHY_ANEG_ADV_100H_ | PHY_ANEG_ADV_100F_ | PHY_SELECTOR_FIELD_));

	/* Enable Auto-Negotiation and
	 * restart Auto-Negotiate.
	 */ 
	smsc911x_set_MIIPHY_reg(1,PHY_BCR, (PHY_BCR_AUTO_NEG_ENABLE_ | PHY_BCR_RESTART_AUTO_NEG_));	

	/* Use 5 seconds as timeout for the auto negotiation. */
	smsc911x_start_elapsed_timer(&tmr,5000);
	
	do {
		status = smsc911x_get_MIIPHY_reg(1,PHY_BSR);

		if(status == 0) {
			printf(NAME_OF_NETDEVICE ": PHY read error while auto negotiating.\n");
			return;
		}
		if(status & PHY_BSR_LINK_STATUS_) {
			if(EtherVerbose & SHOW_DRIVER_DEBUG) {
				printf(NAME_OF_NETDEVICE ": PHY initialized. Link established. \n");
			}
			return;
		}

	} while(!smsc911x_msec_elapsed(&tmr));

	printf(NAME_OF_NETDEVICE ": Auto negotiation timed out.\n");
}

/* smsc911x_set_mac_address():
 * This function is only used by "smsc911x_init()",
 * to encapsule the setting of the mac address. 
 * Returns -1 to indicate an error and 0 to indicate success.
 */
int 
smsc911x_set_mac_address(unsigned char *BinEnetAddr)
{
	unsigned char *m = &BinEnetAddr[0];
	unsigned long addr_high, addr_low;

	/* See if a mac adress was given. */
	if((m[0] | m[1] | m[2] | m[3] | m[4] |m[5]) != 0) {

		addr_low = m[0] | (m[1] << 8) | (m[2] << 16) | (m[3] << 24);
		addr_high = m[4] | (m[5] << 8);

		smsc911x_set_MAC_CSR(ADDRL,addr_low);
		smsc911x_set_MAC_CSR(ADDRH,addr_high);
		return(0);
	}
	return(-1);
}

/* smsc911x_enable():
 * This function is only used by "smsc911x_init()",
 * to encapsule the Tx Rx activation and some settings
 * that should be done after the resets.
 */
void 
smsc911x_enable(void)
{
	/* Set TX FIFO Size. (HW_CFG_TX_FIF_SZ_ = 1000) */
	SMSC911X_SET_REG_BITS(HW_CFG,(8<<16));

	/* Set "Store and Foreward" for data transmission. */
	SMSC911X_SET_REG_BITS(HW_CFG,HW_CFG_SF_);

	/* Set the "General Purpose Timer Pre-Load" 
	 * to "0010011100010000 b".
	 */
	SMSC911X_SET_REG(GPT_CFG,10000);

	/* Start the "GP Timer". */
	SMSC911X_SET_REG_BITS(GPT_CFG, GPT_CFG_TIMER_EN_);

	/* Disable all interrupts. */
	SMSC911X_CLR_REG_BITS(INT_EN,INT_EN_SW_INT_EN_);
	SMSC911X_CLR_REG_BITS(INT_EN,INT_EN_TXSTOP_INT_EN_);
	SMSC911X_CLR_REG_BITS(INT_EN,INT_EN_RXSTOP_INT_EN_);
	SMSC911X_CLR_REG_BITS(INT_EN,INT_EN_RXDFH_INT_EN_);
	SMSC911X_CLR_REG_BITS(INT_EN,INT_EN_TIOC_INT_EN_);
	SMSC911X_CLR_REG_BITS(INT_EN,INT_EN_RXD_INT_EN_);
	SMSC911X_CLR_REG_BITS(INT_EN,INT_EN_GPT_INT_EN_);
	SMSC911X_CLR_REG_BITS(INT_EN,INT_EN_PHY_INT_EN_);
	SMSC911X_CLR_REG_BITS(INT_EN,INT_EN_PME_INT_EN_);
	SMSC911X_CLR_REG_BITS(INT_EN,INT_EN_TXSO_EN_);
	SMSC911X_CLR_REG_BITS(INT_EN,INT_EN_RWT_EN_);
	SMSC911X_CLR_REG_BITS(INT_EN,INT_EN_RXE_EN_);
	SMSC911X_CLR_REG_BITS(INT_EN,INT_EN_TXE_EN_);
	SMSC911X_CLR_REG_BITS(INT_EN,INT_EN_TDFU_EN_);
	SMSC911X_CLR_REG_BITS(INT_EN,INT_EN_TDFO_EN_);
	SMSC911X_CLR_REG_BITS(INT_EN,INT_EN_TDFA_EN_);
	SMSC911X_CLR_REG_BITS(INT_EN,INT_EN_TSFF_EN_);
	SMSC911X_CLR_REG_BITS(INT_EN,INT_EN_TSFL_EN_);
	SMSC911X_CLR_REG_BITS(INT_EN,INT_EN_RXDF_EN_);
	SMSC911X_CLR_REG_BITS(INT_EN,INT_EN_RDFL_EN_);
	SMSC911X_CLR_REG_BITS(INT_EN,INT_EN_RSFF_EN_);
	SMSC911X_CLR_REG_BITS(INT_EN,INT_EN_RSFL_EN_);
	SMSC911X_CLR_REG_BITS(INT_EN,INT_EN_GPIO2_INT_);
	SMSC911X_CLR_REG_BITS(INT_EN,INT_EN_GPIO1_INT_);
	SMSC911X_CLR_REG_BITS(INT_EN,INT_EN_GPIO0_INT_);

	/* Enable the MAC Flow Control function.
	 * Set the value to be used in the PAUSE TIME field in the control frame.
	 * Bits can only be set, when "FLOW_FCBSY_" is cleared.
	 */
	if(smsc911x_get_MAC_CSR(FLOW) & FLOW_FCBSY_) {
		printf(NAME_OF_NETDEVICE 
			": Reset error when setting MAC Flow Control.\n");
		return;
	}

	/* flow controll settings. */

	/* Set the value to be used in the PAUSE TIME field in the control frame. */
	/* Enable the MAC flow control function. */
	smsc911x_set_MAC_CSR(FLOW,(FLOW_FCPT_ | FLOW_FCEN_));

	/* Set "Flow Control on Any Frame"
	 * Setting this bit overrides bits [3:1] of this register.
	 * (Namely: AFC_CFG_FCMULT_, AFC_CFG_FCBRD_ and AFC_CFG_FCADD_)
	 */
	SMSC911X_SET_REG_BITS(AFC_CFG,AFC_CFG_FCANY_);

	/* Set "Backpressure Duration". 
	 * AFC_CFG_BACK_DUR_ = 0111b
	 */
	SMSC911X_SET_REG_BITS(AFC_CFG,0x70);

	/* Set "Automatic Flow Control Low Level". 
	 * AFC_CFG_AFC_LO_ = 00101000b
	 */
	SMSC911X_SET_REG_BITS(AFC_CFG,0x2800);

	/* Set "Automatic Flow Control High Level". 
	 * AFC_CFG_AFC_HI_ = 0x01010000b
	 */
	SMSC911X_SET_REG_BITS(AFC_CFG,0x500000);

	/* End of flow controll settings. */


	/* Enable LEDs 1,2 and 3. */
	SMSC911X_SET_REG_BITS(GPIO_CFG,GPIO_CFG_LED3_EN_);
	SMSC911X_SET_REG_BITS(GPIO_CFG,GPIO_CFG_LED2_EN_);
	SMSC911X_SET_REG_BITS(GPIO_CFG,GPIO_CFG_LED1_EN_);

	/* Set "GPIO Buffer Type" 0,1 and 2 to push/pull driver. */
	SMSC911X_SET_REG_BITS(GPIO_CFG,GPIO_CFG_GPIOBUF2_);
	SMSC911X_SET_REG_BITS(GPIO_CFG,GPIO_CFG_GPIOBUF1_);
	SMSC911X_SET_REG_BITS(GPIO_CFG,GPIO_CFG_GPIOBUF0_);

	/* Enable Transmitter. */
	SMSC911X_SET_REG_BITS(TX_CFG,TX_CFG_TX_ON_);

	/* All RX configurations remain at their default (all 0). */

	/* Enable Transmitter and Receiver between PHY and MAC.
	 * MAC_CR_HBDIS_ unknown (reserved in datasheet), but also
	 * used in the simple driver, when initializing the TX.
	 */
	smsc911x_set_MAC_CSR(MAC_CR,MAC_CR_TXEN_ | MAC_CR_RXEN_ | MAC_CR_HBDIS_);
}

/* smsc911x_init():
 * Start the driver.
 * Return value -1 indicates error, value 0 indicates sucess.
 */
int 
smsc911x_init(void)
{
	if(EtherVerbose & SHOW_DRIVER_DEBUG) {
		printf(NAME_OF_NETDEVICE ": Beginn initializing.\n");
	}

	/* After Power on, wake up or when returning from a power saving
	 * state, the LAN911x must be read at least once to
	 * make write opperations possible.
	 */

	/* Try to find LAN chip ID. */
	if (!smsc911x_verify_chip_ID()) {
		
		if(EtherVerbose & SHOW_DRIVER_DEBUG) {
			printf(NAME_OF_NETDEVICE": LAN controller SMSC911x found.\n");
		}

		/* Reset the network controller. */
		smsc911x_reset();

		/* Configure the PHY and initialize the link state. */
		smsc911x_PHY_reset();
		smsc911x_PHY_configure();

		/* BinEnetAddr is defined in "ether.h" */
		if(smsc911x_set_mac_address(&BinEnetAddr[0]) == 0) {

			/* Enable Tx and Rx. */
			smsc911x_enable();

			if(EtherVerbose & SHOW_DRIVER_DEBUG) {
				printf(NAME_OF_NETDEVICE ": LAN controller init finished.\n");
			}
			return(0);
		}
		else {
			printf(NAME_OF_NETDEVICE ": No valid mac address in config.\n");
		}
	}
	return(-1);
}

/* write_data_to_TX_FIFO_DATA():
 * This function is only used by "smsc911x_send()" to get the packet data
 * into the TX_FIFO_DATA register for transmission.
 */
void 
write_data_to_TX_FIFO_DATA(unsigned long *dataBuf,unsigned long ulCount)
{
	unsigned long *localDataBuf = dataBuf;

	while(ulCount) {
		SMSC911X_SET_REG(TX_DATA_FIFO,*localDataBuf++);
		ulCount--;
	}
}

/* smsc911x_send():
 * Used to send packages from the packet buffer onto the net.
 * Returns 0 when send was a success and -1 when an error has occured.
 */
int 
smsc911x_send(unsigned long *packet,int length)
{
	struct elapsed_tmr tmr;
	int timeout = 1;
	unsigned long freeFIFOspace = 0;
	unsigned long statusTX = 0;

	/* Check incomming packet. */
	if(packet == 0) {
		printf(NAME_OF_NETDEVICE ": Send was called without data.\n");
		return(-1);
	}
	if(length == 0) {
		printf(NAME_OF_NETDEVICE ": Send was called with data length \"0\".\n");
		return(-1);	
	}	

	/* Check available TX FIFO Space*/
	freeFIFOspace = SMSC911X_GET_REG(TX_FIFO_INF);
	freeFIFOspace = freeFIFOspace & TX_FIFO_INF_TDFREE_;
	if(freeFIFOspace < TX_FIFO_LOW_THRESHOLD) {
		printf(NAME_OF_NETDEVICE ": Remaining TX FIFO space to small. Remaining: %ld. \n", freeFIFOspace);
		return(-1);
	}

	/* Place TX COMMAND "A" into TX_DATA_FIFO puffer. */
	SMSC911X_SET_REG(TX_DATA_FIFO,(TX_CMD_A_FIRST_SEG_ | TX_CMD_A_LAST_SEG_ | length));

	/* Place TX COMMAND "B" into TX_DATA_FIFO puffer. */
	SMSC911X_SET_REG(TX_DATA_FIFO,length);

	/* Write the pakage to the TX_DATA_FIFO. */
	write_data_to_TX_FIFO_DATA(((unsigned long *)(packet)),((length + 3) / 4));

	/* Wait till all has been sent out of TX_DATA_FIFO.
	 * When sending is finished, the status is written and TX_FIFO_INF_TSUSED_
	 * indicates how much space of TX_STATUS_FIFO is used.
	 * A 5 second timeout is used.
	 */
	smsc911x_start_elapsed_timer(&tmr,5000);
	do {
		if((SMSC911X_GET_REG(TX_FIFO_INF) & TX_FIFO_INF_TSUSED_) != 0){
			timeout = 0;
			break;
		}
	}while(!smsc911x_msec_elapsed(&tmr));
	if(timeout == 1) {
		printf(NAME_OF_NETDEVICE ": Timeout at package sending.\n");
		return(-1);
	}

	/* Read the status of the successsfull sending.
	 * Transmitting is stopped, when status register overflows.
	 * Status is taken out of FIFO register, when read.
	 */
	statusTX = SMSC911X_GET_REG(TX_STATUS_FIFO);

	/* Take out everything that is not needed, from the status.
	 * TX_STS_TAG		-not used
	 * TX_STS_NO_CARR	-has no meaning when using full duplex
	 */
	statusTX &= ~(TX_STS_TAG_ | TX_STS_NO_CARR_);

	/* Check status for errors. */
	if(!statusTX) {
		return(0);
	}
	else {
		if(statusTX & TX_STS_LOC_) {
			printf(NAME_OF_NETDEVICE ": Loss of Carrier while sending data package.\n");
		}
		if(statusTX & TX_STS_LATE_COLL_) {
			printf(NAME_OF_NETDEVICE ": Sending data package aborted after the collision window 64 bytes.\n");
		}
		if(statusTX & TX_STS_MANY_COLL_) {
			printf(NAME_OF_NETDEVICE ": Sending data package aborted after 16 collisions.\n");
		}
		else {
			if(statusTX & TX_STS_COLL_CNT_) {
				printf(NAME_OF_NETDEVICE ": Data package was sent after %i collisions.\n", (statusTX & TX_STS_COLL_CNT_) >> 3);
				return(0);
			}
		}
		if(statusTX & TX_STS_MANY_DEFER_) {
			printf(NAME_OF_NETDEVICE ": Sending data package aborted a deferral of over 24288 bit times.\n");
		}
		if(statusTX & TX_STS_UNDERRUN_) {
			printf(NAME_OF_NETDEVICE ": Sending data package aborted because of an underrun condition of the TX_DATA_FIFO.\n");
		}
		if(statusTX & TX_STS_DEFERRED_) {
			printf(NAME_OF_NETDEVICE ": Sending data package was deferred.\n");
		}
	}
	return(-1);
}

/* smsc911x_package_arrived():
 * Return 0 when no packet has been received and the length of the data
 * (given in bytes) when a packet has been received.
 */
unsigned long
smsc911x_package_arrived(void)
{
	unsigned long length = 0;

	/* Get the used space inside RX_Status_FIFO (given in amount of unsigned longs).
	 * There can only be something, if a packet was received and its
	 * status was written there.
	 */
	if(SMSC911X_GET_REG(RX_FIFO_INF) & RX_FIFO_INF_RXSUSED_) {

		/* Read status without taking it out of the FIFO register by using the PEEK entry. */
		length = (SMSC911X_GET_REG(RX_STATUS_FIFO_PEEK) & RX_STS_PAKTLENGTH_) >> 16 ;
	}
	return(length);
}

/* smsc911x_receive():
 * Get the received package into the rx buffer.
 * Return 1 on success and -1 when an error has occured.
 */
int 
smsc911x_receive(unsigned long *dataBuffer,unsigned long length)
{
	unsigned long *localDataBuffer = dataBuffer;
	unsigned long lenghtInUL = 0;
	unsigned long status = 0;
	unsigned long deleteBuffer = 0;

	/* Get the complete receive status of this packet out of RX_STATUS_FIFO. */
	status = SMSC911X_GET_REG(RX_STATUS_FIFO);

	/* Find out the length of the buffer (given in Bytes) in unsigned longs. */
	lenghtInUL = (length + 3) / 4;

	/* Only read the packet, if no error has occured. */
	if(!(status & RX_STS_ES_)) {
		
		/* Get the packet from the chip and put it into the software buffer. */
		while(lenghtInUL) {
			*localDataBuffer++ = SMSC911X_GET_REG(RX_DATA_FIFO);
			lenghtInUL--;
		}
		return(1);
	}

	/* When an error has occured, the packet has to be taken out of the RX_DATA_FIFO
	 * but must not be used. */
	else {
		/* Get the packet from the chip and put it into the delete buffer. */
		while(lenghtInUL) {
			deleteBuffer = SMSC911X_GET_REG(RX_DATA_FIFO);
			deleteBuffer = deleteBuffer;
			lenghtInUL--;
		}

		/* Error notification. */
		printf(NAME_OF_NETDEVICE ": Bad packet was dropped." );
		
		if(status & RX_STS_RUNT_FRAME_) {
			printf(" Frame was prematurely terminated.\n");
		}
		if(status & RX_STS_FRAME_TO_LONG_) {
			printf(" Frame length exeeded the Ethernet specifictation.\n");
		}
		if(status & RX_STS_COLLISION_SEEN_) {
			printf(" A late collission has occured.\n");
		}
		if(status & RX_STS_CRC_ERROR_) {
			printf(" CRC error detected.\n");
		}

		/* Other Status. */
		if(status & ~(RX_STS_RUNT_FRAME_ | RX_STS_FRAME_TO_LONG_ | RX_STS_COLLISION_SEEN_ | RX_STS_CRC_ERROR_)) {
			printf("Status: %x \n", status );
		}

	}
	return(-1);
}

/* smsc911x_enablePromiscuousReception():
 * Activate PromiscuousReception.
 * When this mode is active, networkpackets are not filtered
 *.by their MAC address.
 */
void
smsc911x_enablePromiscuousReception(void)
{
		smsc911x_set_MAC_CSR(MAC_CR,(smsc911x_get_MAC_CSR(MAC_CR) | MAC_CR_PRMS_));
}

/* smsc911x_disablePromiscuousReception():
 */
void
smsc911x_disablePromiscuousReception(void)
{
	smsc911x_set_MAC_CSR(MAC_CR,(smsc911x_get_MAC_CSR(MAC_CR) & ~MAC_CR_PRMS_));
}

